# Copyright 2020 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
from quantrocket import __version__
from quantrocket.houston import houston
from quantrocket.cli.utils.output import json_to_cli

def get_version(detail=False):
    """
    Show the QuantRocket version number.

    Parameters
    ----------
    detail : bool
        if True, show the services version number and
        also the version number of the client library making
        this API call. Default is to only show the services
        version number, which is the main QuantRocket version
        number.

    Returns
    -------
    str or dict
        services version number, or dict of services and client
        version numbers
    """

    response = houston.get("/version")
    houston.raise_for_status_with_json(response)
    version = response.json()
    if detail:
        version["client"] = __version__
        return version
    else:
        return version["services"]

def _cli_get_version(detail=False):
    if detail:
        return json_to_cli(get_version, detail=detail)
    else:
        return get_version(), 0



# Copyright 2019 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import os
import time
import itertools
import tempfile
import requests
from quantrocket.master import download_master_file
from quantrocket.exceptions import ParameterError, NoHistoricalData, NoRealtimeData
from quantrocket.history import (
    download_history_file,
    get_db_config as get_history_db_config,
    list_databases as list_history_databases)
from quantrocket.realtime import (
    download_market_data_file,
    get_db_config as get_realtime_db_config,
    list_databases as list_realtime_databases)
from quantrocket.zipline import (
    list_bundles,
    get_bundle_config,
    download_bundle_file)

TMP_DIR = os.environ.get("QUANTROCKET_TMP_DIR", tempfile.gettempdir())

def get_prices(codes, start_date=None, end_date=None,
               universes=None, sids=None,
               exclude_universes=None, exclude_sids=None,
               times=None, fields=None,
               timezone=None, infer_timezone=None,
               cont_fut=None, data_frequency=None):
    """
    Query one or more history databases, real-time aggregate databases,
    and/or Zipline bundles and load prices into a DataFrame.

    For bar sizes smaller than 1-day, the resulting DataFrame will have a MultiIndex
    with levels (Field, Date, Time). For bar sizes of 1-day or larger, the MultiIndex
    will have levels (Field, Date).

    Parameters
    ----------
    codes : str or list of str, required
        the code(s) of one or more databases to query. If multiple databases
        are specified, they must have the same bar size. List databases in order of
        priority (highest priority first). If multiple databases provide the same
        field for the same sid on the same datetime, the first database's value will
        be used.

    start_date : str (YYYY-MM-DD), optional
        limit to data on or after this date

    end_date : str (YYYY-MM-DD), optional
        limit to data on or before this date

    universes : list of str, optional
        limit to these universes (default is to return all securities in database)

    sids : list of str, optional
        limit to these sids

    exclude_universes : list of str, optional
        exclude these universes

    exclude_sids : list of str, optional
        exclude these sids

    times: list of str (HH:MM:SS), optional
        limit to these times, specified in the timezone of the relevant exchange. See
        additional information in the Notes section regarding the timezone to use.

    fields : list of str, optional
        only return these fields. (If querying multiple databases that have different fields,
        provide the complete list of desired fields; only the supported fields for each
        database will be queried.)

    timezone : str, optional
        convert timestamps to this timezone, for example America/New_York (see
        `pytz.all_timezones` for choices); ignored for non-intraday bar sizes

    infer_timezone : bool
        infer the timezone from the securities master Timezone field; defaults to
        True if using intraday bars and no `timezone` specified; ignored for
        non-intraday bars, or if `timezone` is passed

    cont_fut : str
        stitch futures into continuous contracts using this method (default is not
        to stitch together). Only applicable to history databases. Possible choices:
        concat

    data_frequency : str
        for Zipline bundles, whether to query minute or daily data. If omitted,
        defaults to minute data for minute bundles and to daily data for daily bundles.
        This parameter only needs to be set to request daily data from a minute bundle.
        Possible choices: daily, minute (or aliases d, m).

    Returns
    -------
    DataFrame
        a MultiIndex (Field, Date) or (Field, Date, Time) DataFrame of prices

    Notes
    -----
    The `times` parameter, if provided, is applied differently for history databases and
    Zipline bundles vs real-time aggregate databases. For history databases and Zipline
    bundles, the parameter is applied when querying the database. For real-time aggregate
    databases, the parameter is not applied when querying the database; rather, all available
    times are retrieved and the `times` filter is applied to the resulting DataFrame after
    casting it to the appropriate timezone (as inferred from the securities master Timezone
    field or as explicitly specified with the `timezone` parameter). The rationale for this
    behavior is that history databases and Zipline bundles store intraday data in the timezone
    of the relevant exchange whereas real-time aggregate databases store data in UTC. By
    applying the `times` filter as described, users can specify the times in the timezone of
    the relevant exchange for both types of databases.

    Examples
    --------
    Load intraday prices:

    >>> prices = get_prices('stk-sample-5min', fields=["Close", "Volume"])
    >>> prices.head()
                                Sid     	FIBBG1	FIBBG2
    Field	Date	        Time
    Close	2017-07-26      09:30:00	153.62	2715.0
                                09:35:00	153.46	2730.0
                                09:40:00	153.21	2725.0
                                09:45:00	153.28	2725.0
                                09:50:00	153.18	2725.0

    Isolate the closes:

    >>> closes = prices.loc["Close"]
    >>> closes.head()
                Sid	        FIBBG1	FIBBG2
    Date        Time
    2017-07-26	09:30:00	153.62	2715.0
                09:35:00	153.46	2730.0
                09:40:00	153.21	2725.0
                09:45:00	153.28	2725.0
                09:50:00	153.18	2725.0

    Isolate the 15:45:00 prices:

    >>> session_closes = closes.xs("15:45:00", level="Time")
    >>> session_closes.head()
        Sid	FIBBG1	FIBBG2
    Date
    2017-07-26	153.29	2700.00
    2017-07-27 	150.10	2660.00
    2017-07-28	149.43	2650.02
    2017-07-31 	148.99	2650.34
    2017-08-01 	149.72	2675.50
    """
    # Import pandas lazily since it can take a moment to import
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("pandas must be installed to use this function")

    try:
        import pytz
    except ImportError:
        raise ImportError("pytz must be installed to use this function")

    if timezone and timezone not in pytz.all_timezones:
        raise ParameterError(
            "invalid timezone: {0} (see `pytz.all_timezones` for choices)".format(
                timezone))

    dbs = codes
    if not isinstance(dbs, (list, tuple)):
        dbs = [dbs]

    fields = fields or []
    if not isinstance(fields, (list, tuple)):
        fields = [fields]

    # separate history dbs from Zipline bundles from realtime dbs; in case one or
    # more of the services is not running, we print a warning and try the other
    # services
    try:
        history_dbs = set(list_history_databases())
    except requests.HTTPError as e:
        if e.response.status_code == 502:
            import warnings
            warnings.warn(
                f"Error while checking if {', '.join(dbs)} is a history database, "
                f"will assume it's not. Error was: {e}", RuntimeWarning)
            history_dbs = set()
        else:
            raise

    try:
        realtime_dbs = list_realtime_databases()
    except requests.HTTPError as e:
        if e.response.status_code == 502:
            import warnings
            warnings.warn(
                f"Error while checking if {', '.join(dbs)} is a realtime database, "
                f"will assume it's not. Error was: {e}", RuntimeWarning)
            realtime_dbs = {}
            realtime_agg_dbs = set()
        else:
            raise
    else:
        realtime_agg_dbs = set(itertools.chain(*realtime_dbs.values()))

    try:
        zipline_bundles = set(list_bundles())
    except requests.HTTPError as e:
        if e.response.status_code == 502:
            import warnings
            warnings.warn(
                f"Error while checking if {', '.join(dbs)} is a Zipline bundle, "
                f"will assume it's not. Error was: {e}", RuntimeWarning)
            zipline_bundles = set()
        else:
            raise

    history_dbs.intersection_update(set(dbs))
    realtime_agg_dbs.intersection_update(set(dbs))
    zipline_bundles.intersection_update(set(dbs))

    unknown_dbs = set(dbs) - history_dbs - realtime_agg_dbs - zipline_bundles

    if unknown_dbs:
        tick_dbs = set(realtime_dbs.keys()).intersection(unknown_dbs)
        # Improve error message if possible
        if tick_dbs:
            raise ParameterError("{} is a real-time tick database, only history databases or "
                                 "real-time aggregate databases are supported".format(
                                     ", ".join(tick_dbs)))
        raise ParameterError(
            "no history or real-time aggregate databases or Zipline bundles called {}".format(
                ", ".join(unknown_dbs)))

    db_bar_sizes = set()
    db_bar_sizes_parsed = set()
    history_db_fields = {}
    realtime_db_fields = {}
    zipline_bundle_fields = {}

    for db in history_dbs:
        db_config = get_history_db_config(db)
        bar_size = db_config.get("bar_size")
        db_bar_sizes.add(bar_size)
        # to validate uniform bar sizes, we need to parse them in case dbs
        # store different but equivalent timedelta strings. History db
        # strings may need massaging to be parsable.
        if bar_size.endswith("s"):
            # strip s from secs, mins, hours to get valid pandas timedelta
            bar_size = bar_size[:-1]
        elif bar_size == "1 week":
            bar_size = "7 day"
        elif bar_size == "1 month":
            bar_size = "30 day"
        db_bar_sizes_parsed.add(pd.Timedelta(bar_size))
        history_db_fields[db] = db_config.get("fields", [])

    for db in realtime_agg_dbs:
        db_config = get_realtime_db_config(db)
        bar_size = db_config.get("bar_size")
        db_bar_sizes.add(bar_size)
        db_bar_sizes_parsed.add(pd.Timedelta(bar_size))
        realtime_db_fields[db] = db_config.get("fields", [])

    for db in zipline_bundles:
        # look up bundle data_frequency if not specified
        if not data_frequency:
            bundle_config = get_bundle_config(db)
            data_frequency = bundle_config["data_frequency"]

        if data_frequency in ("daily", "d"):
            db_bar_sizes.add("1 day")
            db_bar_sizes_parsed.add(pd.Timedelta("1 day"))
        elif data_frequency in ("minute", "m"):
            db_bar_sizes.add("1 min")
            db_bar_sizes_parsed.add(pd.Timedelta("1 min"))
        else:
            raise ParameterError("invalid data_frequency: {}".format(data_frequency))
        zipline_bundle_fields[db] = ["Open", "High", "Low", "Close", "Volume"]

    if len(db_bar_sizes_parsed) > 1:
        raise ParameterError(
            "all databases must contain same bar size but {0} have different "
            "bar sizes: {1}".format(", ".join(dbs), ", ".join(db_bar_sizes))
        )

    all_prices = []

    for db in dbs:

        if db in history_dbs:
            # different DBs might support different fields so only request the
            # subset of supported fields
            fields_for_db = set(fields).intersection(set(history_db_fields[db]))

            kwargs = dict(
                start_date=start_date,
                end_date=end_date,
                universes=universes,
                sids=sids,
                exclude_universes=exclude_universes,
                exclude_sids=exclude_sids,
                times=times,
                cont_fut=cont_fut,
                fields=list(fields_for_db),
            )

            tmp_filepath = "{dir}{sep}history.{db}.{pid}.{time}.csv".format(
                dir=TMP_DIR, sep=os.path.sep, db=db, pid=os.getpid(), time=time.time())

            try:
                download_history_file(db, tmp_filepath, **kwargs)
            except NoHistoricalData:
                # don't complain about NoHistoricalData if we're checking
                # multiple databases, unless none of them have data
                if len(dbs) == 1:
                    raise
                else:
                    continue

            prices = pd.read_csv(tmp_filepath)
            prices = prices.pivot(index="Sid", columns="Date").T
            prices.index.set_names(["Field", "Date"], inplace=True)
            all_prices.append(prices)

            os.remove(tmp_filepath)

        if db in realtime_agg_dbs:

            fields_for_db = set(fields).intersection(set(realtime_db_fields[db]))

            kwargs = dict(
                start_date=start_date,
                end_date=end_date,
                universes=universes,
                sids=sids,
                exclude_universes=exclude_universes,
                exclude_sids=exclude_sids,
                fields=list(fields_for_db))

            tmp_filepath = "{dir}{sep}realtime.{db}.{pid}.{time}.csv".format(
                dir=TMP_DIR, sep=os.path.sep, db=db, pid=os.getpid(), time=time.time())

            try:
                download_market_data_file(db, tmp_filepath, **kwargs)
            except NoRealtimeData as e:
                # don't complain about NoRealtimeData if we're checking
                # multiple databases, unless none of them have data
                if len(dbs) == 1:
                    raise
                else:
                    continue

            prices = pd.read_csv(tmp_filepath)
            prices = prices.pivot(index="Sid", columns="Date").T
            prices.index.set_names(["Field", "Date"], inplace=True)
            all_prices.append(prices)

            os.remove(tmp_filepath)

        if db in zipline_bundles:

            fields_for_db = set(fields).intersection(set(zipline_bundle_fields[db]))

            kwargs = dict(
                start_date=start_date,
                end_date=end_date,
                universes=universes,
                sids=sids,
                exclude_universes=exclude_universes,
                exclude_sids=exclude_sids,
                times=times,
                data_frequency=data_frequency,
                fields=list(fields_for_db))

            tmp_filepath = "{dir}{sep}zipline.{db}.{pid}.{time}.csv".format(
                dir=TMP_DIR, sep=os.path.sep, db=db, pid=os.getpid(), time=time.time())

            try:
                download_bundle_file(db, tmp_filepath, **kwargs)
            except NoHistoricalData as e:
                # don't complain about NoHistoricalData if we're checking
                # multiple databases, unless none of them have data
                if len(dbs) == 1:
                    raise
                else:
                    continue

            prices = pd.read_csv(tmp_filepath, index_col=["Field", "Date"])
            prices.columns.name = "Sid"
            all_prices.append(prices)

            os.remove(tmp_filepath)

    # complain if multiple dbs and none had data
    if len(dbs) > 1 and not all_prices:
        raise NoHistoricalData("no price data matches the query parameters in any of {0}".format(
            ", ".join(dbs)
        ))

    prices = None
    for _prices in all_prices:
        if prices is None:
            prices = _prices
        else:
            prices = prices.combine_first(_prices)

    is_intraday = list(db_bar_sizes_parsed)[0] < pd.Timedelta("1 day")

    # For intraday dbs, infer timezone from securities master
    if is_intraday and not timezone and infer_timezone is not False:

        infer_timezone = True
        sids = list(prices.columns)

        f = six.StringIO()
        download_master_file(
            f,
            sids=sids,
            fields="Timezone")
        securities = pd.read_csv(f, index_col="Sid")

        timezones = securities.Timezone.unique()

        if len(timezones) > 1:
            raise ParameterError(
                "cannot infer timezone because multiple timezones are present "
                "in data, please specify timezone explicitly (timezones: {0})".format(
                    ", ".join(timezones)))

        timezone = timezones[0]

    if is_intraday:
        dates = pd.to_datetime(prices.index.get_level_values("Date"), utc=True)

        if timezone:
            dates = dates.tz_convert(timezone)
    else:
        # use .str[:10] because the format might be 2020-04-05 (history dbs)
        # or 2020-04-05T00:00:00-00 (realtime aggregate dbs)
        dates = pd.to_datetime(
            prices.index.get_level_values("Date").str[:10])

    prices.index = pd.MultiIndex.from_arrays((
        prices.index.get_level_values("Field"),
        dates
        ), names=("Field", "Date"))

    # Split date and time
    dts = prices.index.get_level_values("Date")
    dates = pd.to_datetime(dts.date).tz_localize(None) # drop tz-aware in Date index
    prices.index = pd.MultiIndex.from_arrays(
        (prices.index.get_level_values("Field"),
         dates,
         dts.strftime("%H:%M:%S")),
        names=["Field", "Date", "Time"]
    )

    # Align dates if there are any duplicate. Explanation: Suppose there are
    # two timezones represented in the data (e.g. history db in security
    # timezone vs real-time db in UTC). After parsing these dates into a
    # common timezone, they will align properly, but we pivoted before
    # parsing the dates (for performance reasons), so they may not be
    # aligned. Thus we need to dedupe the index. NOTE: first() is slow
    # when prices contains categorical/object dtypes (which can happen
    # with custom fundamental databases, for example), so only do this
    # if we have to. nth(0) would be an alternative to first() but it
    # doesn't ignore nans (nth()'s dropna param doesn't help here), so
    # a universal solution is as yet elusive.
    if prices.index.duplicated().any():
        prices = prices.groupby(prices.index).first()
        prices.index = pd.MultiIndex.from_tuples(prices.index)
        prices.index.set_names(["Field", "Date", "Time"], inplace=True)

    # Fill missing dates and times so that each field has the
    # same set of dates and times, for easier vectorized operations.
    # Example implications for intraday dbs:
    # - if history is retrieved intraday, this ensures that today will have NaN
    #   entries for future times
    # - early close dates will have a full set of times, with NaNs after the
    #   early close
    unique_fields = prices.index.get_level_values("Field").unique()
    unique_dates = prices.index.get_level_values("Date").unique()
    unique_times = prices.index.get_level_values("Time").unique()
    interpolated_index = None
    for field in unique_fields:
        field_idx = pd.MultiIndex.from_product([[field], unique_dates, unique_times]).sort_values()
        if interpolated_index is None:
            interpolated_index = field_idx
        else:
            interpolated_index = interpolated_index.append(field_idx)

    prices = prices.reindex(interpolated_index)
    prices.index.set_names(["Field", "Date", "Time"], inplace=True)

    # Drop time if not intraday
    if not is_intraday:
        prices.index = prices.index.droplevel("Time")
        return prices

    # Apply times filter if needed (see Notes in docstring)
    if times and realtime_agg_dbs:
        if not isinstance(times, (list, tuple)):
            times = [times]
        prices = prices.loc[prices.index.get_level_values("Time").isin(times)]

    return prices

def get_prices_reindexed_like(reindex_like, codes, fields=None,
                              shift=1, ffill=True, lookback_window=10,
                              agg="last",
                              timezone=None, infer_timezone=None,
                              times=None, cont_fut=None,
                              data_frequency=None):
    """
    Return a multiindex (Field, Date) DataFrame of prices for one or more history
    databases, real-time aggregate databases, or Zipline bundles, reindexed to match
    the index (dates) and columns (sids) of `reindex_like`.

    Prices are loaded with `quantrocket.get_prices` and shifted forward one period
    (configurable with the `shift` parameter) to avoid lookahead bias. In the case
    of sparse data, values are then forward-filled by default (configurable with the
    `ffill` parameter).

    The queried databases need not contain price data. This function can be used to query
    custom history databases containing any kind of data.

    Pay attention to the `lookback_window` parameter, which controls how much back data
    in advance of the input DataFrame's start date to load. For example, if the input
    DataFrame contains daily data and you are querying quarterly fundamental data,
    the `lookback_window` must extend far enough back in time to access the most recent
    quarterly value prior to the input DataFrame's start date. Setting `lookback_window`
    too low will result in leading NaNs in the resulting DataFrame. Setting it too high
    is okay but will load a larger amount of data into memory.

    If you query an intraday database, it will be treated as a daily database. Specifically,
    the intraday values will be aggregated (using the method specified by the `agg`
    parameter) to produce a single value per day.

    If the input DataFrame has levels in the index other than Date (for example, Time, in
    the case of an intraday input DataFrame), the queried values will be broadcast across
    the additional levels of the index.

    Parameters
    ----------
    reindex_like : DataFrame, required
        a DataFrame with dates for the index (or, if DataFrame has a MultiIndex, with
        dates for one level of the index) and sids for the columns, to which the shape
        of the resulting DataFrame will be conformed.

    codes : str or list of str, required
        the code(s) of one or more databases to query. If multiple databases
        are specified, they must have the same bar size. List databases in order of
        priority (highest priority first). If multiple databases provide the same
        field for the same sid on the same datetime, the first database's value will
        be used.

    fields : list of str, optional
        only return these fields. (If querying multiple databases that have different fields,
        provide the complete list of desired fields; only the supported fields for each
        database will be queried.)

    shift : int, optional
        number of periods (in the date index) to shift the resulting data forward to
        avoid lookahead bias. Default is 1. Shifting one period implies that data
        timestamped to a particular date is available and actionable on the following
        date.

    ffill : bool
        forward-fill values in the resulting DataFrame so that each date reflects
        the latest available value as of that date. If False, values appear only
        on the first date they were available, followed by NaNs. Default True.

    lookback_window : int, optional
        how many calendar days of back data prior to the reindex_like start date
        should be loaded, to ensure an adequate cushion of data is available before
        shifting. Default is 10. Sparse data such as fundamentals will require a
        higher value.

    times: list of str (HH:MM:SS), optional
        limit to these times, specified in the timezone of the relevant exchange. Only
        applicable to querying intraday databases. See additional information in the
        Notes section of `quantrocket.get_prices` regarding the timezone to use.

    timezone : str, optional
        convert timestamps to this timezone, for example America/New_York (see
        `pytz.all_timezones` for choices); ignored for non-intraday databases

    infer_timezone : bool
        infer the timezone from the securities master Timezone field; defaults to
        True if querying intraday databases and no `timezone` specified; ignored for
        non-intraday databases, or if `timezone` is passed

    agg : str or function, optional
        when querying intraday databases, how to aggregate each day's intraday values to
        produce a single value per day. Default is "last", meaning use the last non-null
        value of each day. This parameter is passed directly to the pandas `agg` function.
        Example choices include "last", "first", "min", "max", "mean", "sum", etc. See
        https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.core.groupby.DataFrameGroupBy.aggregate.html
        for more info. Note that aggregation occurs after the `times` filters are applied.

    cont_fut : str
        stitch futures into continuous contracts using this method (default is not
        to stitch together). Only applicable to history databases. Possible choices:
        concat

    data_frequency : str
        for Zipline bundles, whether to query minute or daily data. If omitted,
        defaults to minute data for minute bundles and to daily data for daily bundles.
        This parameter only needs to be set to request daily data from a minute bundle.
        Possible choices: daily, minute (or aliases d, m).

    Returns
    -------
    DataFrame
        a MultiIndex DataFrame of prices shaped like reindex_like

    Examples
    --------
    Use a DataFrame of closing prices to query a DataFrame of fundamentals in a
    database called "custom-fundamentals". Since the fundamental data is sparse,
    we specify a lookback window of 180 days to ensure that a previous value can
    be forward-filled into the resulting DataFrame:

    >>> closes = prices.loc["Close"]
    >>> fundamentals = get_prices_reindexed_like(
        closes, "custom-fundamentals", fields="Revenue",
        lookback_window=180)
    >>> revenues = fundamentals.loc["Revenue"]
    """
    try:
        import pandas as pd
    except ImportError:
        raise ImportError("pandas must be installed to use this function")

    index_levels = reindex_like.index.names

    if "Date" not in index_levels:
        raise ParameterError(
            "reindex_like must have index called 'Date', but has {0}".format(
                ",".join([str(name) for name in index_levels])))

    reindex_like_dates = reindex_like.index.get_level_values("Date")

    is_multiindex = len(reindex_like.index.names) > 1

    if not hasattr(reindex_like_dates, "date"):
        raise ParameterError("reindex_like must have a DatetimeIndex")

    sids = list(reindex_like.columns)
    start_date = reindex_like_dates.min().date()
    start_date -= pd.Timedelta(days=lookback_window)
    start_date = start_date.isoformat()
    end_date = reindex_like_dates.max().date().isoformat()

    prices = get_prices(
        codes,
        sids=sids,
        fields=fields,
        start_date=start_date,
        end_date=end_date,
        timezone=timezone,
        times=times,
        cont_fut=cont_fut,
        data_frequency=data_frequency)

    all_fields = {}

    # default values are set in the signature for docstring purposes, but we
    # also set them here in case Nones were passed
    if shift is None:
        shift = 1
    if lookback_window is None:
        lookback_window = 10
    if agg is None:
        agg = "last"

    fields = prices.index.get_level_values("Field").unique()

    for field in fields:

        prices_for_field = prices.loc[field]

        # For intraday databases, drop Time and aggregate per date
        if "Time" in prices_for_field.index.names:
            prices_for_field.index = prices_for_field.index.get_level_values("Date")
            prices_for_field = prices_for_field.groupby(prices_for_field.index).agg(agg)

        # get_prices returns tz-naive dates, localize to match reindex_like
        prices_for_field = prices_for_field.tz_localize(reindex_like_dates.tz)

        # union indexes and reindex in case there are any price dates not in
        # reindex_like or vice versa
        unioned_idx = reindex_like_dates.union(prices_for_field.index).drop_duplicates()
        prices_for_field = prices_for_field.reindex(index=unioned_idx)

        if ffill:
            prices_for_field = prices_for_field.fillna(method="ffill")

        # shift forward to avoid lookahead bias
        prices_for_field = prices_for_field.shift(shift)

        # Reindex to requested shape
        if is_multiindex:
            prices_for_field = prices_for_field.reindex(
                index=reindex_like.index, level="Date").reindex(
                    columns=reindex_like.columns)
        else:
            prices_for_field = prices_for_field.reindex(
                index=reindex_like.index, columns=reindex_like.columns)

        all_fields[field] = prices_for_field

    names = ["Field"] + reindex_like.index.names
    prices = pd.concat(all_fields, names=names)

    return prices



# Copyright 2017 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

def add_subparser(subparsers):
    _parser = subparsers.add_parser("flightlog", description="QuantRocket logging service CLI", help="Monitor and download logs")
    _subparsers = _parser.add_subparsers(title="subcommands", dest="subcommand")
    _subparsers.required = True

    examples = """
Stream application logs, `tail -f` style.

Examples:

Stream application logs:

    quantrocket flightlog stream

Stream detailed logs:

    quantrocket flightlog stream --detail
    """
    parser = _subparsers.add_parser(
        "stream",
        help="stream application logs, `tail -f` style",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "-d", "--detail",
        action="store_true",
        help="show detailed logs from logspout, otherwise show log messages from flightlog only")
    parser.add_argument(
        "--hist",
        type=int,
        metavar="NUM_LINES",
        help="number of log lines to show right away (ignored if showing detailed logs)")
    parser.add_argument(
        "--nocolor",
        action="store_false",
        dest="color",
        help="don't colorize the logs")
    parser.set_defaults(func="quantrocket.flightlog._cli_stream_logs")

    examples = """
Download the logfile.

Examples:

Download application logs:

    quantrocket flightlog get app.log

Download detailed logs:

    quantrocket flightlog get --detail sys.log

Download detailed logs for the history service:

    quantrocket flightlog get --detail --match quantrocket_history sys.log
    """
    parser = _subparsers.add_parser(
        "get",
        help="download the logfile",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "outfile",
        metavar="OUTFILE",
        help="filename to write the logfile to")
    parser.add_argument(
        "-d", "--detail",
        action="store_true",
        help="download detailed logs from the logspout service, otherwise download "
        "the standard logs from the flightlog service")
    parser.add_argument(
        "-m", "--match",
        metavar="PATTERN",
        help="filter the logfile to lines containing this string")
    parser.set_defaults(func="quantrocket.flightlog._cli_download_logfile")

    examples = r"""
Wait for a message to appear in the logs.

Searches can be performed against the standard or detailed log file.
When searching the detailed logs, note that the log file uses the
syslog format, which differs from the format used when streaming
detailed logs. Download the detailed log file to see the exact format
your search will run against.

Examples:

Wait up to 10 minutes for a message to appear indicating that data
ingestion has finished:

    quantrocket flightlog wait '[usstock-1min] Completed ingesting data' --timeout 10m

Using a regular expression, wait up to 1 hour for a message to appear
indicating that data collection has finished:

    quantrocket flightlog wait '\[usstock-1d\] Collected [0-9]+ monthly files' --regex --timeout 1h
    """
    parser = _subparsers.add_parser(
        "wait",
        help="wait for a message to appear in the logs",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "message",
        help="the log message to search for")
    parser.add_argument(
        "-r", "--regex",
        action="store_true",
        help="if True, treat the `message` argument as a regular expression "
        "(default is to treat it as a plain string)")
    parser.add_argument(
        "-d", "--detail",
        action="store_true",
        help="if True, search the detailed logs from the logspout service (default "
        "is to search the standard logs from the flightlog service)")
    parser.add_argument(
        "--tail",
        type=int,
        help="search the most recent N lines of the logs in addition to searching "
        "future logs (default is to only search future logs)")
    parser.add_argument(
        "--timeout",
        help="fail if the message is not found after this much time (use Pandas "
        "timedelta string, e.g. 30sec or 5min or 2h; default is to wait "
        "indefinitely)")
    parser.set_defaults(func="quantrocket.flightlog._cli_wait_for_message")

    examples = """
Log a message.

Examples:

Log a message under the name "myapp":

    quantrocket flightlog log "this is a test" --name myapp --level INFO

Log the output from another command:

    quantrocket account balance --below-cushion 0.02 | quantrocket flightlog log --name quantrocket.account --level CRITICAL
    """
    parser = _subparsers.add_parser(
        "log",
        help="log a message",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "msg",
        nargs="?",
        default="-",
        help="the message to be logged")
    parser.add_argument(
        "-l", "--level",
        default="INFO",
        metavar="LEVEL",
        choices=("DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"),
        help="the log level for the message. Possible choices: %(choices)s")
    parser.add_argument(
        "-n", "--name",
        dest="logger_name",
        default="quantrocket.cli",
        help="the logger name")
    parser.set_defaults(func="quantrocket.flightlog._cli_log_message")

    examples = """
Set or show the flightlog timezone.

Examples:

Set the flightlog timezone to America/New_York:

    quantrocket flightlog timezone America/New_York

Show the current flightlog timezone:

    quantrocket flightlog timezone
    """
    parser = _subparsers.add_parser(
        "timezone",
        help="set or show the flightlog timezone",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "tz",
        nargs="?",
        metavar="TZ",
        help="the timezone to set (pass a partial timezone string such as 'newyork' "
        "or 'europe' to see close matches, or pass '?' to see all choices)")
    parser.set_defaults(func="quantrocket.flightlog._cli_get_or_set_timezone")

    examples = """
Set or show the Papertrail log configuration.

See http://qrok.it/h/pt to learn more.

Examples:

Set the Papertrail host and port to log to:

    quantrocket flightlog papertrail --host logs.papertrailapp.com --port 55555

Show the current papertrail config:

    quantrocket flightlog papertrail
    """
    parser = _subparsers.add_parser(
        "papertrail",
        help="set or show the Papertrail log configuration",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "--host",
        metavar="HOST",
        help="the Papertrail host to log to")
    parser.add_argument(
        "--port",
        metavar="PORT",
        type=int,
        help="the Papertrail port to log to")
    parser.set_defaults(func="quantrocket.flightlog._cli_get_or_set_papertrail_config")



# Copyright 2017 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import six
import requests
from quantrocket.houston import houston
from quantrocket.exceptions import NoHistoricalData
from quantrocket.cli.utils.output import json_to_cli
from quantrocket.cli.utils.stream import to_bytes
from quantrocket.cli.utils.files import write_response_to_filepath_or_buffer
from quantrocket.cli.utils.parse import dict_strs_to_dict, dict_to_dict_strs
from quantrocket.utils.warn import deprecated_replaced_by

def create_usstock_bundle(code, sids=None, universes=None, free=False, data_frequency=None):
    """
    Create a Zipline bundle for US stocks.

    This function defines the bundle parameters but does not ingest the actual
    data. To ingest the data, see `ingest_bundle`.

    Parameters
    ----------
    code : str, required
        the code to assign to the bundle (lowercase alphanumerics and hyphens only)

    sids : list of str, optional
        limit to these sids (only supported for minute data bundles)

    universes : list of str, optional
        limit to these universes (only supported for minute data bundles)

    free : bool
        limit to free sample data

    data_frequency : str, optional
         whether to collect minute data (which also includes daily data) or only
         daily data. Default is minute data. Possible choices: daily, minute (or
         aliases d, m)

    Returns
    -------
    dict
        status message

    Examples
    --------
    Create a minute data bundle for all US stocks:

    >>> create_usstock_bundle("usstock-1min")

    Create a bundle for daily data only:

    >>> create_usstock_bundle("usstock-1d", data_frequency="daily")

    Create a minute data bundle based on a universe:

    >>> create_usstock_bundle("usstock-tech-1min", universes="us-tech")

    Create a minute data bundle of free sample data:

    >>> create_usstock_bundle("usstock-free-1min", free=True)
    """
    params = {}
    params["ingest_type"] = "usstock"
    if sids:
        params["sids"] = sids
    if universes:
        params["universes"] = universes
    if free:
        params["free"] = free
    if data_frequency:
        params["data_frequency"] = data_frequency

    response = houston.put("/zipline/bundles/{}".format(code), params=params)

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_create_usstock_bundle(*args, **kwargs):
    return json_to_cli(create_usstock_bundle, *args, **kwargs)

def create_bundle_from_db(code, from_db, calendar,
                         start_date=None, end_date=None,
                         universes=None, sids=None,
                        exclude_universes=None, exclude_sids=None,
                        fields=None):
    """
    Create a Zipline bundle from a history database or real-time aggregate
    database.

    You can ingest 1-minute or 1-day databases.

    This function defines the bundle parameters but does not ingest the actual
    data. To ingest the data, see `ingest_bundle`.

    Parameters
    ----------
    code : str, required
        the code to assign to the bundle (lowercase alphanumerics and hyphens only)

    from_db : str, required
        the code of a history database or real-time aggregate database to ingest

    calendar : str, required
        the name of the calendar to use with this bundle (provide '?' or
        any invalid calendar name to see available choices)

    start_date : str (YYYY-MM-DD), required
        limit to historical data on or after this date. This parameter is required
        and also determines the default start date for backtests and queries.

    end_date : str (YYYY-MM-DD), optional
        limit to historical data on or before this date

    universes : list of str, optional
        limit to these universes

    sids : list of str, optional
        limit to these sids

    exclude_universes : list of str, optional
        exclude these universes

    exclude_sids : list of str, optional
        exclude these sids

    fields : dict, optional
        mapping of Zipline fields (open, high, low, close, volume) to
        db fields. Defaults to mapping Zipline 'open' to db 'Open', etc.

    Returns
    -------
    dict
        status message

    Examples
    --------
    Create a bundle from a history database called "es-fut-1min" and name
    it like the history database:

    >>> create_bundle_from_db("es-fut-1min", from_db="es-fut-1min", calendar="us_futures")

    Create a bundle named "usa-stk-1min-2017" for ingesting a single year of
    US 1-minute stock data from a history database called "usa-stk-1min":

    >>> create_bundle_from_db("usa-stk-1min-2017", from_db="usa-stk-1min",
    >>>                      calendar="XNYS",
    >>>                      start_date="2017-01-01", end_date="2017-12-31")

    Create a bundle from a real-time aggregate database and specify how to map
    Zipline fields to the database fields:

    >>> create_bundle_from_db("free-stk-1min", from_db="free-stk-tick-1min",
    >>>                       calendar="XNYS", start_date="2020-06-01",
    >>>                       fields={
    >>>                           "close": "LastPriceClose",
    >>>                           "open": "LastPriceOpen",
    >>>                           "high": "LastPriceHigh",
    >>>                           "low": "LastPriceLow",
    >>>                           "volume": "VolumeClose"})
    """
    params = {}
    params["ingest_type"] = "from_db"
    params["from_db"] = from_db
    params["calendar"] = calendar
    if start_date:
        params["start_date"] = start_date
    if end_date:
        params["end_date"] = end_date
    if universes:
        params["universes"] = universes
    if sids:
        params["sids"] = sids
    if exclude_universes:
        params["exclude_universes"] = exclude_universes
    if exclude_sids:
        params["exclude_sids"] = exclude_sids
    if fields:
        params["fields"] = dict_to_dict_strs(fields)

    response = houston.put("/zipline/bundles/{}".format(code), params=params)

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_create_bundle_from_db(*args, **kwargs):
    fields = kwargs.get("fields", None)
    if fields:
        kwargs["fields"] = dict_strs_to_dict(*fields)
    return json_to_cli(create_bundle_from_db, *args, **kwargs)

def ingest_bundle(code, sids=None, universes=None):
    """
    Ingest data into a previously defined bundle.

    Parameters
    ----------
    code : str, required
        the bundle code

    sids : list of str, optional
        limit to these sids, overriding stored config

    universes : list of str, optional
        limit to these universes, overriding stored config

    Returns
    -------
    dict
        status message

    Examples
    --------
    Ingest data into a bundle called usstock-1min:

    >>> ingest_bundle("usstock-1min")
    """
    params = {}
    if sids:
        params["sids"] = sids
    if universes:
        params["universes"] = universes

    response = houston.post("/zipline/ingestions/{}".format(code), params=params)

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_ingest_bundle(*args, **kwargs):
    return json_to_cli(ingest_bundle, *args, **kwargs)

def list_bundles():
    """
    List available data bundles and whether data has been
    ingested into them.

    Returns
    -------
    dict
        data bundles and whether they have data (True indicates
        data, False indicates config only)
    """
    response = houston.get("/zipline/bundles")

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_list_bundles(*args, **kwargs):
    return json_to_cli(list_bundles, *args, **kwargs)

def get_bundle_config(code):
    """
    Return the configuration of a bundle.

    Parameters
    ----------
    code : str, required
        the bundle code

    Returns
    -------
    dict
        config
    """
    response = houston.get(f"/zipline/bundles/config/{code}")

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_get_bundle_config(*args, **kwargs):
    return json_to_cli(get_bundle_config, *args, **kwargs)

def drop_bundle(code, confirm_by_typing_bundle_code_again=None):
    """
    Delete a bundle.

    Parameters
    ----------
    code : str, required
        the bundle code

    confirm_by_typing_bundle_code_again : str, required
       enter the bundle code again to confirm you want to drop the bundle, its config,
       and all its data

    Returns
    -------
    dict
        status message

    Examples
    --------
    Delete a bundle called 'es-fut-1min':

    >>> drop_bundle("es-fut-1min")
    """
    params = {}
    if confirm_by_typing_bundle_code_again:
        params["confirm_by_typing_bundle_code_again"] = confirm_by_typing_bundle_code_again

    response = houston.delete("/zipline/bundles/{}".format(code), params=params, timeout=6*60*60)

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_drop_bundle(*args, **kwargs):
    return json_to_cli(drop_bundle, *args, **kwargs)

def get_default_bundle():
    """
    Return the current default bundle, if any.

    Returns
    -------
    dict
        default bundle
    """
    response = houston.get("/zipline/config")
    houston.raise_for_status_with_json(response)
    # It's possible to get a 204 empty response
    if not response.content:
        return {}
    return response.json()

def set_default_bundle(bundle):
    """
    Set the default bundle to use for backtesting and trading.

    Setting a default bundle is a convenience and is optional. It
    can be overridden by manually specifying a bundle when backtesting
    or trading.

    Parameters
    ----------
    bundle : str, required
        the bundle code

    Returns
    -------
    dict
        status message
    """
    data = {
        "default_bundle": bundle
    }
    response = houston.put("/zipline/config", data=data)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_get_or_set_default_bundle(bundle=None, *args, **kwargs):
    if bundle:
        return json_to_cli(set_default_bundle, bundle, *args, **kwargs)
    else:
        return json_to_cli(get_default_bundle, *args, **kwargs)

def download_bundle_file(code, filepath_or_buffer=None,
                         start_date=None, end_date=None,
                         data_frequency=None,
                         universes=None, sids=None,
                         exclude_universes=None, exclude_sids=None,
                         times=None, fields=None):
    """
    Query minute or daily data from a Zipline bundle and download to a CSV file.

    Parameters
    ----------
    code : str, required
        the bundle code

    filepath_or_buffer : str or file-like object
        filepath to write the data to, or file-like object (defaults to stdout)

    start_date : str (YYYY-MM-DD), optional
        limit to history on or after this date

    end_date : str (YYYY-MM-DD), optional
        limit to history on or before this date

    data_frequency : str, optional
        whether to query minute or daily data. If omitted, defaults to minute data
        for minute bundles and to daily data for daily bundles. This parameter
        only needs to be set to request daily data from a minute bundle. Possible
        choices: daily, minute (or aliases d, m).

    universes : list of str, optional
        limit to these universes

    sids : list of str, optional
        limit to these sids

    exclude_universes : list of str, optional
        exclude these universes

    exclude_sids : list of str, optional
        exclude these sids

    times: list of str (HH:MM:SS), optional
        limit to these times

    fields : list of str, optional
        only return these fields (pass ['?'] or any invalid fieldname to see
        available fields)

    Returns
    -------
    None

    Examples
    --------
    Load minute prices into pandas:

    >>> download_bundle_file("usstock-1min", sids=["FIBBG12345"])
    >>> prices = pd.read_csv(f, parse_dates=["Date"], index_col=["Field","Date"])

    Isolate fields with .loc:

    >>> closes = prices.loc["Close"]

    See Also
    --------
    quantrocket.get_prices : load prices into a DataFrame
    """
    params = {}
    if start_date:
        params["start_date"] = start_date
    if end_date:
        params["end_date"] = end_date
    if data_frequency:
        params["data_frequency"] = data_frequency
    if universes:
        params["universes"] = universes
    if sids:
        params["sids"] = sids
    if exclude_universes:
        params["exclude_universes"] = exclude_universes
    if exclude_sids:
        params["exclude_sids"] = exclude_sids
    if times:
        params["times"] = times
    if fields:
        params["fields"] = fields

    response = houston.get("/zipline/bundles/data/{0}.csv".format(code), params=params,
                           timeout=60*30)

    try:
        houston.raise_for_status_with_json(response)
    except requests.HTTPError as e:
        # Raise a dedicated exception
        # This endpoint returns "no minute|daily data match the query parameters"
        if "data match the query parameters" in repr(e).lower():
            raise NoHistoricalData(e)
        raise

    filepath_or_buffer = filepath_or_buffer or sys.stdout

    write_response_to_filepath_or_buffer(filepath_or_buffer, response)

def _cli_download_bundle_file(*args, **kwargs):
    return json_to_cli(download_bundle_file, *args, **kwargs)

@deprecated_replaced_by(download_bundle_file)
def download_minute_file(*args, **kwargs):
    kwargs["data_frequency"] = "minute"
    return download_bundle_file(*args, **kwargs)

def backtest(strategy, data_frequency=None, capital_base=None, bundle=None,
             start_date=None, end_date=None, progress=None, filepath_or_buffer=None):
    """
    Backtest a Zipline strategy and write the test results to a CSV file.

    The CSV result file contains several DataFrames stacked into one: the Zipline performance
    results, plus the extracted returns, transactions, positions, and benchmark returns from those
    results.

    Parameters
    ----------
    strategy : str, required
        the strategy to run (strategy filename without extension)

    data_frequency : str, optional
        the data frequency of the simulation. Possible choices: daily, minute
        (or aliases d, m). Default is minute.

    capital_base : float, optional
        the starting capital for the simulation (default is 1e6 (1 million))

    bundle : str, optional
        the data bundle to use for the simulation. If omitted, the default bundle (if set)
        is used.

    start_date : str (YYYY-MM-DD), optional
        the start date of the simulation (defaults to the bundle start date)

    end_date : str (YYYY-MM-DD), optional
        the end date of the simulation (defaults to today)

    progress : str, optional
        log backtest progress at this interval (use a pandas offset alias,
        for example "D" for daily, "W" for weeky, "M" for monthly,
        "A" for annually)

    filepath_or_buffer : str, optional
        the location to write the output file (omit to write to stdout)

    Returns
    -------
    None

    Examples
    --------
    Run a backtest defined in momentum-pipeline.py and save to CSV,
    logging backtest progress at weekly intervals.

    >>> backtest("momentum-pipeline", bundle="my-bundle",
                 start_date="2015-02-04", end_date="2015-12-31",
                 progress="W",
                 filepath_or_buffer="momentum_pipeline_results.csv")

    Get a pyfolio tear sheet from the results:

    >>> import pyfolio as pf
    >>> pf.from_zipline_csv("momentum_pipeline_results.csv")
    """
    params = {}
    if data_frequency:
        params["data_frequency"] = data_frequency
    if capital_base:
        params["capital_base"] = capital_base
    if bundle:
        params["bundle"] = bundle
    if start_date:
        params["start_date"] = start_date
    if end_date:
        params["end_date"] = end_date
    if progress:
        params["progress"] = progress

    response = houston.post("/zipline/backtests/{0}".format(strategy), params=params, timeout=60*60*96)

    houston.raise_for_status_with_json(response)

    filepath_or_buffer = filepath_or_buffer or sys.stdout
    write_response_to_filepath_or_buffer(filepath_or_buffer, response)

def _cli_backtest(*args, **kwargs):
    return json_to_cli(backtest, *args, **kwargs)

def create_tearsheet(infilepath_or_buffer, outfilepath_or_buffer=None):
    """
    Create a pyfolio PDF tear sheet from a Zipline backtest result.

    Parameters
    ----------
    infilepath_or_buffer : str, required
        the CSV file from a Zipline backtest (specify '-' to read file from stdin)

    outfilepath_or_buffer : str or file-like, optional
        the location to write the pyfolio tear sheet (write to stdout if omitted)

    Returns
    -------
    None
    """
    url = "/zipline/tearsheets"
    # Pyfolio can take a long time
    timeout = 60*60*5

    if infilepath_or_buffer == "-":
        infilepath_or_buffer = sys.stdin.buffer if six.PY3 else sys.stdin
        response = houston.post(url, data=infilepath_or_buffer, timeout=timeout)

    elif infilepath_or_buffer and hasattr(infilepath_or_buffer, "read"):
        if infilepath_or_buffer.seekable():
            infilepath_or_buffer.seek(0)
        response = houston.post(url, data=infilepath_or_buffer, timeout=timeout)

    else:
        with open(infilepath_or_buffer, "rb") as f:
            response = houston.post(url, data=f, timeout=timeout)

    houston.raise_for_status_with_json(response)

    outfilepath_or_buffer = outfilepath_or_buffer or sys.stdout
    write_response_to_filepath_or_buffer(outfilepath_or_buffer, response)

def _cli_create_tearsheet(*args, **kwargs):
    return json_to_cli(create_tearsheet, *args, **kwargs)

def trade(strategy, bundle=None, account=None, data_frequency=None):
    """
    Trade a Zipline strategy.

    Parameters
    ----------
    strategy : str, required
        the strategy to run (strategy filename without extension)

    bundle : str, optional
        the data bundle to use. If omitted, the default bundle (if set)
        is used.

    account : str, optional
        the account to run the strategy in. Only required
        if the strategy is allocated to more than one
        account in quantrocket.zipline.allocations.yml.

    data_frequency : str, optional
        the data frequency to use. Possible choices: daily, minute
        (or aliases d, m). Default is minute.

    Returns
    -------
    None

    Examples
    --------
    Trade a strategy defined in momentum-pipeline.py:

    >>> trade("momentum-pipeline", bundle="my-bundle")
    """
    params = {}
    if bundle:
        params["bundle"] = bundle
    if account:
        params["account"] = account
    if data_frequency:
        params["data_frequency"] = data_frequency

    response = houston.post("/zipline/trade/{0}".format(strategy), params=params)

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_trade(*args, **kwargs):
    return json_to_cli(trade, *args, **kwargs)

def list_active_strategies():
    """
    List actively trading Zipline strategies.

    Returns
    -------
    dict
        dict of account: strategies
    """
    response = houston.get("/zipline/trade")

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_list_active_strategies(*args, **kwargs):
    return json_to_cli(list_active_strategies, *args, **kwargs)

def cancel_strategies(strategies=None, accounts=None, cancel_all=False):
    """
    Cancel actively trading strategies.

    Parameters
    ----------
    strategies : list of str, optional
        limit to these strategies

    accounts : list of str, optional
        limit to these accounts

    cancel_all : bool
        cancel all actively trading strategies

    Returns
    -------
    dict
        dict of actively trading strategies after canceling

    Examples
    --------
    Cancel a single strategy:

    >>> cancel_strategies(strategies="momentum-pipeline")

    Cancel all strategies:

    >>> cancel_strategies(cancel_all=True)
    """
    params = {}
    if strategies:
        params["strategies"] = strategies
    if accounts:
        params["accounts"] = accounts
    if cancel_all:
        params["cancel_all"] = cancel_all

    response = houston.delete("/zipline/trade", params=params)

    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_cancel_strategies(*args, **kwargs):
    return json_to_cli(cancel_strategies, *args, **kwargs)

class ZiplineBacktestResult(object):
    """
    Convenience class for parsing a CSV result file from a Zipline backtest
    into a variety of useful DataFrames, which can be passed to pyfolio or
    inspected by the user.

    Examples
    --------
    Run a Zipline backtest and parse the CSV results:

    >>> f = io.StringIO()
    >>> run_algorithm("momentum_pipeline.py",
              bundle="etf-sampler-1d",
              start="2015-02-04",
              end="2015-12-31",
              filepath_or_buffer=f)
    >>> zipline_result = ZiplineBacktestResult.from_csv(f)

    The ZiplineBacktestResult object contains returns, positions, transactions,
    benchmark_returns, and the performance DataFrame.

    >>> print(zipline_result.returns.head())
    >>> print(zipline_result.positions.head())
    >>> print(zipline_result.transactions.head())
    >>> print(zipline_result.benchmark_returns.head())
    >>> print(zipline_result.perf.head())

    The outputs are ready to be passed to pyfolio:

    >>> pf.create_full_tear_sheet(
            zipline_result.returns,
            positions=zipline_result.positions,
            transactions=zipline_result.transactions,
            benchmark_rets=zipline_result.benchmark_returns)
    """

    def __init__(self):
        self.returns = None
        self.positions = None
        self.transactions = None
        self.benchmark_returns = None
        self.perf = None

    @classmethod
    def from_csv(cls, filepath_or_buffer):

        # Import pandas lazily since it can take a moment to import
        try:
            import pandas as pd
        except ImportError:
            raise ImportError("pandas must be installed to use ZiplineBacktestResult")

        zipline_result = cls()

        results = pd.read_csv(
            filepath_or_buffer,
            parse_dates=["date"],
            index_col=["dataframe", "index", "date", "column"])["value"]

        # Extract returns
        returns = results.loc["returns"].unstack()
        is_tz_naive = returns.index.droplevel(0).tz is None
        returns.index = returns.index.droplevel(0)
        if is_tz_naive:
            returns.index = returns.index.tz_localize("UTC")
        else:
            returns.index = returns.index.tz_convert("UTC")
        zipline_result.returns = returns["returns"].astype(float)

        # Extract positions
        if "positions" in results.index.get_level_values("dataframe"):
            positions = results.loc["positions"].unstack()
            positions.index = positions.index.droplevel(0)
            if is_tz_naive:
                positions.index = positions.index.tz_localize("UTC")
            else:
                positions.index = positions.index.tz_convert("UTC")
            zipline_result.positions = positions.astype(float)

        # Extract transactions
        transactions = results.loc["transactions"].unstack()
        transactions.index = transactions.index.droplevel(0)
        if is_tz_naive:
            transactions.index = transactions.index.tz_localize("UTC")
        else:
            transactions.index = transactions.index.tz_convert("UTC")
        zipline_result.transactions = transactions.apply(pd.to_numeric, errors='ignore')

        # Extract benchmark returns
        if "benchmark" in results.index.get_level_values("dataframe"):
            benchmark_returns = results.loc["benchmark"].unstack()
            benchmark_returns.index = benchmark_returns.index.droplevel(0)
            if is_tz_naive:
                benchmark_returns.index = benchmark_returns.index.tz_localize("UTC")
            else:
                benchmark_returns.index = benchmark_returns.index.tz_convert("UTC")
            zipline_result.benchmark_returns = benchmark_returns["benchmark"].astype(float)

        # Extract performance dataframe
        perf = results.loc["perf"].unstack()
        perf.index = perf.index.droplevel(0)
        if is_tz_naive:
            perf.index = perf.index.tz_localize("UTC")
        else:
            perf.index = perf.index.tz_convert("UTC")
        zipline_result.perf = perf.apply(pd.to_numeric, errors='ignore')

        return zipline_result



# Copyright 2018 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse

def add_subparser(subparsers):
    examples = """
Show the QuantRocket version number.

Examples:

Show the version number:

    quantrocket version

Show both the services and client version numbers:

    quantrocket version -d
    """
    parser = subparsers.add_parser(
        "version",
        description="show the QuantRocket version number",
        help="show the QuantRocket version number",
        epilog=examples,
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(
        "-d", "--detail",
        action="store_true",
        help="show the services version number and "
        "also the version number of the client library making "
        "this API call. Default is to only show the services "
        "version number, which is the main QuantRocket version "
        "number.")
    parser.set_defaults(func="quantrocket.version._cli_get_version")



# Copyright 2017 QuantRocket - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import getpass
from quantrocket.houston import houston
from quantrocket.exceptions import DataInsertionError
from quantrocket.cli.utils.output import json_to_cli

def list_databases(services=None, codes=None, detail=False, expand=False):
    """
    List databases.

    Parameters
    ----------
    services : str, optional
        limit to these services

    codes: list of str, optional
        limit to these codes

    detail : bool
        return database statistics (default is to return a
        flat list of database names)

    expand : bool
        expand sharded databases to include individual shards
        (default is to list sharded databases as a single database)

    Returns
    -------
    dict
        dict of lists of databases (one key for PostgreSQL databases and one for
        SQLite databases)

    Examples
    --------
    Load database details in a pandas DataFrame:

    >>> from quantrocket.db import list_databases
    >>> import itertools
    >>> databases = list_databases(detail=True)
    >>> databases = pd.DataFrame.from_records(itertools.chain(databases["sqlite"], databases["postgres"]))
    """
    params = {}
    if services:
        params["services"] = services
    if codes:
        params["codes"] = codes
    if detail:
        params["detail"] = detail
    if expand:
        params["expand"] = expand

    response = houston.get("/db/databases", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_list_databases(*args, **kwargs):
    return json_to_cli(list_databases, *args, **kwargs)

def get_s3_config():
    """
    Return the current S3 configuration, if any.

    See http://qrok.it/h/dbs3 to learn more.

    Returns
    -------
    dict
        configuration details
    """
    response = houston.get("/db/s3config")
    houston.raise_for_status_with_json(response)
    # It's possible to get a 204 empty response
    if not response.content:
        return {}
    return response.json()

def set_s3_config(access_key_id=None, secret_access_key=None, bucket=None, region=None):
    """
    Set AWS S3 configuration for pushing and pulling databases to and from
    S3.

    See http://qrok.it/h/dbs3 to learn more.

    Credentials are encrypted at rest and never leave your deployment.

    Parameters
    ----------
    access_key_id : str, optional
        AWS access key ID

    secret_access_key : str, optional
        AWS secret access key (if omitted and access_key_id is provided,
        will be prompted for secret_access_key)

    bucket : str, optional
        the S3 bucket name to push to/pull from

    region : str, optional
        the AWS region in which to create the bucket (default us-east-1).
        Ignored if the bucket already exists.

    Returns
    -------
    dict
        status message
    """
    if access_key_id and not secret_access_key:
        secret_access_key = getpass.getpass(prompt="Enter AWS Secret Access Key: ")

    data = {}
    if access_key_id:
        data["access_key_id"] = access_key_id
    if secret_access_key:
        data["secret_access_key"] = secret_access_key
    if bucket:
        data["bucket"] = bucket
    if region:
        data["region"] = region

    response = houston.put("/db/s3config", data=data)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_get_or_set_s3_config(access_key_id=None, secret_access_key=None,
                                   bucket=None, region=None, *args, **kwargs):
    if access_key_id or secret_access_key or bucket or region:
        return json_to_cli(set_s3_config, access_key_id, secret_access_key, bucket, region, *args, **kwargs)
    else:
        return json_to_cli(get_s3_config, *args, **kwargs)

def s3_push_databases(services=None, codes=None):
    """
    Push database(s) to Amazon S3.

    See http://qrok.it/h/dbs3 to learn more.

    Parameters
    ----------
    serivces : list of str, optional
        limit to these services

    codes: list of str, optional
        limit to these codes

    Returns
    -------
    json
        status message
    """
    params = {}
    if services:
        params["services"] = services
    if codes:
        params["codes"] = codes
    response = houston.put("/db/s3", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_s3_push_databases(*args, **kwargs):
    return json_to_cli(s3_push_databases, *args, **kwargs)

def s3_pull_databases(services=None, codes=None, force=False):
    """
    Pull database(s) from Amazon S3.

    See http://qrok.it/h/dbs3 to learn more.

    Parameters
    ----------
    serivces : list of str, optional
        limit to these services

    codes: list of str, optional
        limit to these codes

    force: bool
        overwrite existing database if one exists (default is to
        fail if one exists)

    Returns
    -------
    json
        status message
    """
    params = {}
    if services:
        params["services"] = services
    if codes:
        params["codes"] = codes
    if force:
        params["force"] = force
    response = houston.get("/db/s3", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_s3_pull_databases(*args, **kwargs):
    return json_to_cli(s3_pull_databases, *args, **kwargs)

def optimize_databases(services=None, codes=None):
    """
    Optimize databases to improve performance.

    This runs the 'VACUUM' command, which defragments the database and
    reclaims disk space.

    Parameters
    ----------
    serivces : list of str, optional
        limit to these service

    codes: list of str, optional
        limit to these codes

    Returns
    -------
    json
        status message
    """
    params = {}
    if codes:
        params["codes"] = codes
    if services:
        params["services"] = services
    response = houston.post("/db/optimizations", params=params)
    houston.raise_for_status_with_json(response)
    return response.json()

def _cli_optimize_databases(*args, **kwargs):
    return json_to_cli(optimize_databases, *args, **kwargs)

def connect_sqlite(db_path):
    """
    Returns a connection to a SQLite database.

    Parameters
    ----------
    db_path : str, required
        full path to a SQLite database

    Returns
    -------
    sqlalchemy.engine.Engine
        database connection
    """
    try:
        from sqlalchemy import create_engine
    except ImportError:
        raise ValueError(
            "this function requires sqlalchemy and must be run in a QuantRocket container")

    conn = create_engine("sqlite:///{0}".format(db_path),
                         connect_args={"isolation_level": None})
    # Set some speed optimizations
    # Hand off writes to the OS and don't wait
    conn.execute("PRAGMA synchronous = 0")
    # Each page is ~1K; allow ~50MB
    conn.execute("PRAGMA cache_size = 50000")
    # Store temp tables in memory
    conn.execute("PRAGMA temp_store = 2")
    # Wait up to 10 seconds rather than instantly failing on SQLITE_BUSY
    conn.execute("PRAGMA busy_timeout = 10000")

    return conn

def _insert_into(df, table_name, conn, on_conflict):

    import time
    import subprocess
    import os

    temp_table_name = "temp_{0}".format(str(time.time()).replace(".", ""))

    # Get the db path from the engine object
    db_path = conn.url.database

    temp_file_name = "/tmp/sqlite_{}.csv".format(temp_table_name)

    # Cast booleans to ints or they will load into SQLite as strings
    df_bools = df.select_dtypes(['bool'])
    if not df_bools.empty:
        df.loc[:, df_bools.columns] = df_bools.astype(int)

    df.to_csv(temp_file_name, index=False)

    # Close connection to avoid Database Is Locked
    conn.dispose()

    from_cols = [f"NULLIF({col}, '')" for col in df.columns]

    queries = """
PRAGMA busy_timeout = 10000;
.bail on
.mode csv
.import {tempfile} {temptable}
INSERT OR {on_conflict} INTO {table} ({into_cols}) SELECT {from_cols} FROM {temptable};
DROP TABLE {temptable};
    """.format(
        table=table_name,
        on_conflict=on_conflict,
        into_cols=",".join(df.columns),
        from_cols=",".join(from_cols),
        tempfile=temp_file_name,
        temptable=temp_table_name)

    try:
        subprocess.check_output(
            ["sqlite3", db_path],
            input=bytes(queries.encode("utf-8")),
            stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        raise DataInsertionError(e.output)

    os.remove(temp_file_name)

def insert_or_fail(df, table_name, conn):
    """
    Insert a DataFrame into a SQLite database.

    In the case of a duplicate record insertion, the function
    will fail.

    Parameters
    ----------
    df : DataFrame, required
        the DataFrame to insert. All DataFrame columns must
        exist in the destination table. The DataFrame index
        will not be inserted.

    table_name : str, required
        the name of the table to insert the DataFrame into.
        The table must already exist in the database.

    conn : sqlalchemy.engine.Engine, required
        a connection object for the SQLite database

    Returns
    -------
    None

    Raises
    ------
    quantrocket.exceptions.DataInsertionError
        catch-all exception class for errors that occur when writing to the
        SQLite database
    """
    _insert_into(df, table_name, conn, "FAIL")

def insert_or_replace(df, table_name, conn):
    """
    Insert a DataFrame into a SQLite database.

    In the case of a duplicate record insertion, the incoming
    record will replace the existing record.

    Parameters
    ----------
    df : DataFrame, required
        the DataFrame to insert. All DataFrame columns must
        exist in the destination table. The DataFrame index
        will not be inserted.

    table_name : str, required
        the name of the table to insert the DataFrame into.
        The table must already exist in the database.

    conn : sqlalchemy.engine.Engine, required
        a connection object for the SQLite database

    Returns
    -------
    None

    Raises
    ------
    quantrocket.exceptions.DataInsertionError
        catch-all exception class for errors that occur when writing to the
        SQLite database
    """
    _insert_into(df, table_name, conn, "REPLACE")

def insert_or_ignore(df, table_name, conn):
    """
    Insert a DataFrame into a SQLite database.

    In the case of a duplicate record insertion, the incoming
    record will be ignored.

    Parameters
    ----------
    df : DataFrame, required
        the DataFrame to insert. All DataFrame columns must
        exist in the destination table. The DataFrame index
        will not be inserted.

    table_name : str, required
        the name of the table to insert the DataFrame into.
        The table must already exist in the database.

    conn : sqlalchemy.engine.Engine, required
        a connection object for the SQLite database

    Returns
    -------
    None

    Raises
    ------
    quantrocket.exceptions.DataInsertionError
        catch-all exception class for errors that occur when writing to the
        SQLite database
    """
    _insert_into(df, table_name, conn, "IGNORE")



# Copyright 2019 QuantRocket LLC - All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# To run: python -m unittest discover -s tests/ -p test*.py -t .

import unittest
try:
    from unittest.mock import patch
except ImportError:
    # py27
    from mock import patch
import pandas as pd
import pytz
import numpy as np
from quantrocket.master import (
    get_securities,
    get_securities_reindexed_like,
    get_contract_nums_reindexed_like)
from quantrocket.exceptions import ParameterError

class GetSecuritiesTestCase(unittest.TestCase):

    @patch("quantrocket.master.download_master_file")
    def test_pass_args_correctly(self, mock_download_master_file):
        """
        Tests that args are correctly passed to the download_master_file
        function.
        """
        def _mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                          "FI23456"],
                     Symbol=["ABC",
                             "DEF"],
                     Etf=[1,
                          0],
                     Delisted=[0,
                               1],
                     Currency=["USD",
                               "USD"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        mock_download_master_file.side_effect = _mock_download_master_file

        kwargs = dict(
            symbols=["ABC", "DEF"],
            exchanges="XNYS",
            currencies="USD",
            sec_types=["STK", "ETF"],
            universes="my-universe",
            sids=["FI12345", "FI23456"],
            exclude_universes="exclude-me",
            exclude_sids="FI34567",
            exclude_delisted=True,
            exclude_expired=True,
            frontmonth=True,
            vendors=["usstock"],
            fields=["Symbol", "Etf", "Delisted", "Currency"]
        )

        get_securities(**kwargs)

        download_master_file_call = mock_download_master_file.mock_calls[0]
        _, args, kwargs = download_master_file_call
        self.assertListEqual(kwargs["symbols"], ["ABC", "DEF"])
        self.assertEqual(kwargs["exchanges"], "XNYS")
        self.assertEqual(kwargs["currencies"], "USD")
        self.assertListEqual(kwargs["sec_types"], ["STK", "ETF"])
        self.assertListEqual(kwargs["sids"], ["FI12345","FI23456"])
        self.assertEqual(kwargs["universes"], "my-universe")
        self.assertEqual(kwargs["exclude_universes"], "exclude-me")
        self.assertEqual(kwargs["exclude_sids"], "FI34567")
        self.assertTrue(kwargs["exclude_delisted"])
        self.assertTrue(kwargs["exclude_expired"])
        self.assertTrue(kwargs["frontmonth"])
        self.assertListEqual(kwargs["vendors"], ["usstock"])
        self.assertListEqual(kwargs["fields"], ["Symbol", "Etf", "Delisted", "Currency"])

    @patch("quantrocket.master.download_master_file")
    def test_cast_boolean_and_date_fields(self, mock_download_master_file):
        """
        Tests that master fields are correctly cast to boolean and datetime.
        """
        self.maxDiff = None
        def _mock_download_master_file(f, *args, **kwargs):
            stks = [{
                'Sid': 'FIBBG000B9XRY4',
                'Symbol': 'AAPL',
                'Exchange': 'XNAS',
                'Country': 'US',
                'Currency': 'USD',
                'SecType': 'STK',
                'Etf': 0,
                'Timezone': 'America/New_York',
                'Name': 'APPLE INC',
                'PriceMagnifier': 1,
                'Multiplier': 1,
                'Delisted': 0,
                'DateDelisted': None,
                'LastTradeDate': None,
                'RolloverDate': None,
                'alpaca_AssetClass': 'us_equity',
                'alpaca_AssetId': 'b0b6dd9d-8b9b-48a9-ba46-b9d54906e415',
                'alpaca_EasyToBorrow': 1,
                'alpaca_Exchange': 'NASDAQ',
                'alpaca_Marginable': 1,
                'alpaca_Name': 'Apple Inc. Common Stock',
                'alpaca_Shortable': 1,
                'alpaca_Status': 'active',
                'alpaca_Symbol': 'AAPL',
                'alpaca_Tradable': 1,
                'edi_Cik': 320193,
                'edi_CountryInc': 'United States of America',
                'edi_CountryListed': 'United States of America',
                'edi_Currency': 'USD',
                'edi_DateDelisted': None,
                'edi_ExchangeListingStatus': 'Listed',
                'edi_FirstPriceDate': '2017-03-01',
                'edi_GlobalListingStatus': 'Active',
                'edi_Industry': 'Information Technology',
                'edi_IsPrimaryListing': 1,
                'edi_IsoCountryInc': 'US',
                'edi_IsoCountryListed': 'US',
                'edi_IssuerId': 30017,
                'edi_IssuerName': 'Apple Inc',
                'edi_LastPriceDate': '2020-11-04',
                'edi_LocalSymbol': 'AAPL',
                'edi_Mic': 'XNAS',
                'edi_MicSegment': 'XNGS',
                'edi_MicTimezone': 'America/New_York',
                'edi_PreferredName': 'Apple Inc',
                'edi_PrimaryMic': 'XNAS',
                'edi_RecordCreated': '2001-05-05',
                'edi_RecordModified': '2020-11-02 03:37:23',
                'edi_SecId': 33449,
                'edi_SecTypeCode': 'EQS',
                'edi_SecTypeDesc': 'Equity Shares',
                'edi_SecurityDesc': 'Ordinary Shares',
                'edi_Sic': 'Electronic Computers',
                'edi_SicCode': 3571,
                'edi_SicDivision': 'Manufacturing',
                'edi_SicIndustryGroup': 'Computer And Office Equipment',
                'edi_SicMajorGroup': 'Industrial And Commercial Machinery And Computer Equipment',
                'edi_StructureCode': None,
                'edi_StructureDesc': None,
                'ibkr_AggGroup': 1,
                'ibkr_Category': 'Computers',
                'ibkr_ComboLegs': None,
                'ibkr_ConId': 265598,
                'ibkr_ContractMonth': None,
                'ibkr_Currency': 'USD',
                'ibkr_Cusip': None,
                'ibkr_DateDelisted': None,
                'ibkr_Delisted': 0,
                'ibkr_Etf': 0,
                'ibkr_EvMultiplier': 0.0,
                'ibkr_EvRule': None,
                'ibkr_Industry': 'Computers',
                'ibkr_Isin': 'US0378331005',
                'ibkr_LastTradeDate': None,
                'ibkr_LocalSymbol': 'AAPL',
                'ibkr_LongName': 'APPLE INC',
                'ibkr_MarketName': 'NMS',
                'ibkr_MarketRuleIds': '26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26',
                'ibkr_MdSizeMultiplier': 100,
                'ibkr_MinTick': 0.01,
                'ibkr_Multiplier': None,
                'ibkr_PriceMagnifier': 1,
                'ibkr_PrimaryExchange': 'NASDAQ',
                'ibkr_RealExpirationDate': None,
                'ibkr_Right': None,
                'ibkr_SecType': 'STK',
                'ibkr_Sector': 'Technology',
                'ibkr_Strike': 0.0,
                'ibkr_Symbol': 'AAPL',
                'ibkr_Timezone': 'America/New_York',
                'ibkr_TradingClass': 'NMS',
                'ibkr_UnderConId': 0,
                'ibkr_UnderSecType': None,
                'ibkr_UnderSymbol': None,
                'ibkr_ValidExchanges': 'SMART,AMEX,NYSE,CBOE,PHLX,ISE,CHX,ARCA,ISLAND,DRCTEDGE,BEX,BATS,EDGEA,CSFBALGO,JEFFALGO,BYX,IEX,EDGX,FOXRIVER,TPLUS1,NYSENAT,LTSE,MEMX,PSX',
                'sharadar_Category': 'Domestic Common Stock',
                'sharadar_CompanySite': 'http://www.apple.com',
                'sharadar_CountryListed': 'US',
                'sharadar_Currency': 'USD',
                'sharadar_Cusips': 37833100,
                'sharadar_DateDelisted': None,
                'sharadar_Delisted': 0,
                'sharadar_Exchange': 'NASDAQ',
                'sharadar_FamaIndustry': 'Computers',
                'sharadar_FamaSector': None,
                'sharadar_FirstAdded': '2014-09-24',
                'sharadar_FirstPriceDate': '1986-01-01',
                'sharadar_FirstQuarter': '1996-09-30',
                'sharadar_Industry': 'Consumer Electronics',
                'sharadar_LastPriceDate': '2020-11-04',
                'sharadar_LastQuarter': '2020-09-30',
                'sharadar_LastUpdated': '2020-11-04',
                'sharadar_Location': 'California; U.S.A',
                'sharadar_Name': 'Apple Inc',
                'sharadar_Permaticker': 199059,
                'sharadar_RelatedTickers': None,
                'sharadar_ScaleMarketCap': '6 - Mega',
                'sharadar_ScaleRevenue': '6 - Mega',
                'sharadar_SecFilings': 'https://www.sec.gov/cgi-bin/browse-edgar?action=getcompany&CIK=0000320193',
                'sharadar_Sector': 'Technology',
                'sharadar_SicCode': 3571,
                'sharadar_SicIndustry': 'Electronic Computers',
                'sharadar_SicSector': 'Manufacturing',
                'sharadar_Ticker': 'AAPL',
                'usstock_DateDelisted': None,
                'usstock_FirstPriceDate': '2007-01-03',
                'usstock_LastPriceDate': '2020-11-18',
                'usstock_Mic': 'XNAS',
                'usstock_Name': 'APPLE INC',
                'usstock_SecurityType': 'Common Stock',
                'usstock_SecurityType2': 'Common Stock',
                'usstock_Sic': 'Electronic Computers',
                'usstock_SicCode': 3571,
                'usstock_SicDivision': 'Manufacturing',
                'usstock_SicIndustryGroup': 'Computer And Office Equipment',
                'usstock_SicMajorGroup': 'Industrial And Commercial Machinery And Computer Equipment',
                'usstock_Symbol': 'AAPL'},
                {'Sid': 'FIBBG000001PV7',
                'Symbol': 'GOODO',
                'Exchange': 'XNAS',
                'Country': 'US',
                'Currency': 'USD',
                'SecType': 'STK',
                'Etf': 0,
                'Timezone': 'America/New_York',
                'Name': 'GLADSTONE COMMERCIAL COR',
                'PriceMagnifier': 1,
                'Multiplier': 1,
                'Delisted': 1,
                'DateDelisted': '2019-10-25',
                'LastTradeDate': None,
                'RolloverDate': None,
                'edi_Cik': 1234006,
                'edi_CountryInc': 'United States of America',
                'edi_CountryListed': 'United States of America',
                'edi_Currency': 'USD',
                'edi_DateDelisted': '2019-10-25',
                'edi_ExchangeListingStatus': 'Delisted',
                'edi_FirstPriceDate': '2017-03-01',
                'edi_GlobalListingStatus': 'Inactive',
                'edi_Industry': 'Real Estate',
                'edi_IsPrimaryListing': 1,
                'edi_IsoCountryInc': 'US',
                'edi_IsoCountryListed': 'US',
                'edi_IssuerId': 74929,
                'edi_IssuerName': 'Gladstone Commercial Corp',
                'edi_LastPriceDate': '2019-10-25',
                'edi_LocalSymbol': 'GOODO',
                'edi_Mic': 'XNAS',
                'edi_MicSegment': 'XNGS',
                'edi_MicTimezone': 'America/New_York',
                'edi_PreferredName': 'Gladstone Commercial Corp',
                'edi_PrimaryMic': 'XNAS',
                'edi_RecordCreated': '2006-10-26',
                'edi_RecordModified': '2019-11-04 03:52:18',
                'edi_SecId': 426219,
                'edi_SecTypeCode': 'PRF',
                'edi_SecTypeDesc': 'Preference Share',
                'edi_SecurityDesc': '7.50% PRF PERPETUAL USD 25 - Reg S Ser B',
                'edi_Sic': 'Lessors of Real Property, Not Elsewhere Classified',
                'edi_SicCode': 6519,
                'edi_SicDivision': 'Finance, Insurance, And Real Estate',
                'edi_SicIndustryGroup': 'Real Estate Operators (except Developers) And Lessors',
                'edi_SicMajorGroup': 'Real Estate',
                'edi_StructureCode': 'REIT',
                'edi_StructureDesc': 'Real Estate Investment Trust',
                'sharadar_Category': 'Domestic Preferred Stock',
                'sharadar_CompanySite': None,
                'sharadar_CountryListed': 'US',
                'sharadar_Currency': 'USD',
                'sharadar_Cusips': 376536306,
                'sharadar_DateDelisted': '2019-10-25',
                'sharadar_Delisted': 1,
                'sharadar_Exchange': 'NASDAQ',
                'sharadar_FamaIndustry': 'Real Estate',
                'sharadar_FamaSector': None,
                'sharadar_FirstAdded': '2018-12-30',
                'sharadar_FirstPriceDate': '2006-10-25',
                'sharadar_FirstQuarter': None,
                'sharadar_Industry': 'REIT - Diversified',
                'sharadar_LastPriceDate': '2019-10-25',
                'sharadar_LastQuarter': None,
                'sharadar_LastUpdated': '2020-07-01',
                'sharadar_Location': 'Virginia; U.S.A',
                'sharadar_Name': 'Gladstone Commercial Corp',
                'sharadar_Permaticker': 112800,
                'sharadar_RelatedTickers': 'GOODP GOODM GOOD',
                'sharadar_ScaleMarketCap': None,
                'sharadar_ScaleRevenue': None,
                'sharadar_SecFilings': None,
                'sharadar_Sector': 'Real Estate',
                'sharadar_SicCode': 6519,
                'sharadar_SicIndustry': 'Lessors Of Real Property Nec',
                'sharadar_SicSector': 'Finance Insurance And Real Estate',
                'sharadar_Ticker': 'GOODO',
                'usstock_DateDelisted': '2019-10-25',
                'usstock_FirstPriceDate': '2007-01-03',
                'usstock_LastPriceDate': '2019-10-25',
                'usstock_Mic': 'XNAS',
                'usstock_Name': 'GLADSTONE COMMERCIAL COR',
                'usstock_SecurityType': 'PUBLIC',
                'usstock_SecurityType2': 'Preferred Stock',
                'usstock_Sic': 'Lessors of Real Property, Not Elsewhere Classified',
                'usstock_SicCode': 6519,
                'usstock_SicDivision': 'Finance, Insurance, And Real Estate',
                'usstock_SicIndustryGroup': 'Real Estate Operators (except Developers) And Lessors',
                'usstock_SicMajorGroup': 'Real Estate',
                'usstock_Symbol': 'GOODO'},
                {'Sid': 'QF000000000004',
                'Symbol': 'ESH7',
                'Exchange': 'XCME',
                'Country': 'US',
                'Currency': 'USD',
                'SecType': 'FUT',
                'Etf': 0,
                'Timezone': 'America/Chicago',
                'Name': 'E-mini S&P 500',
                'PriceMagnifier': 1,
                'Multiplier': 50,
                'Delisted': 0,
                'DateDelisted': None,
                'LastTradeDate': '2017-03-17',
                'RolloverDate': '2017-03-15',
                'ibkr_AggGroup': 2147483647,
                'ibkr_Category': None,
                'ibkr_ComboLegs': None,
                'ibkr_ConId': 215465490,
                'ibkr_ContractMonth': 201703,
                'ibkr_Currency': 'USD',
                'ibkr_Cusip': None,
                'ibkr_DateDelisted': None,
                'ibkr_Delisted': 0,
                'ibkr_Etf': 0,
                'ibkr_EvMultiplier': 0.0,
                'ibkr_EvRule': None,
                'ibkr_Industry': None,
                'ibkr_Isin': None,
                'ibkr_LastTradeDate': '2017-03-17',
                'ibkr_LocalSymbol': 'ESH7',
                'ibkr_LongName': 'E-mini S&P 500',
                'ibkr_MarketName': 'ES',
                'ibkr_MarketRuleIds': None,
                'ibkr_MdSizeMultiplier': 1,
                'ibkr_MinTick': 0.25,
                'ibkr_Multiplier': 50,
                'ibkr_PriceMagnifier': 1,
                'ibkr_PrimaryExchange': 'GLOBEX',
                'ibkr_RealExpirationDate': None,
                'ibkr_Right': None,
                'ibkr_SecType': 'FUT',
                'ibkr_Sector': None,
                'ibkr_Strike': 0.0,
                'ibkr_Symbol': 'ES',
                'ibkr_Timezone': 'America/Chicago',
                'ibkr_TradingClass': 'ES',
                'ibkr_UnderConId': 11004968,
                'ibkr_UnderSecType': 'IND',
                'ibkr_UnderSymbol': 'ES',
                'ibkr_ValidExchanges': 'GLOBEX'}
            ]
            securities = pd.DataFrame(stks)
            securities.to_csv(f, index=False)
            f.seek(0)

        mock_download_master_file.side_effect = _mock_download_master_file

        with patch("quantrocket.price.download_master_file", new=mock_download_master_file):

            securities = get_securities(sids=["FIBBG000B9XRY4", "FIBBG000001PV7", "QF000000000004"], fields="*")

        self.assertListEqual(
            securities.reset_index().sort_values("Sid").fillna('nan').to_dict(orient="records"),
            [{'Sid': 'FIBBG000001PV7',
            'Symbol': 'GOODO',
            'Exchange': 'XNAS',
            'Country': 'US',
            'Currency': 'USD',
            'SecType': 'STK',
            'Etf': False,
            'Timezone': 'America/New_York',
            'Name': 'GLADSTONE COMMERCIAL COR',
            'PriceMagnifier': 1,
            'Multiplier': 1,
            'Delisted': True,
            'DateDelisted': pd.Timestamp('2019-10-25 00:00:00'),
            'LastTradeDate': 'nan',
            'RolloverDate': 'nan',
            'alpaca_AssetClass': 'nan',
            'alpaca_AssetId': 'nan',
            'alpaca_EasyToBorrow': False,
            'alpaca_Exchange': 'nan',
            'alpaca_Marginable': False,
            'alpaca_Name': 'nan',
            'alpaca_Shortable': False,
            'alpaca_Status': 'nan',
            'alpaca_Symbol': 'nan',
            'alpaca_Tradable': False,
            'edi_Cik': 1234006.0,
            'edi_CountryInc': 'United States of America',
            'edi_CountryListed': 'United States of America',
            'edi_Currency': 'USD',
            'edi_DateDelisted': pd.Timestamp('2019-10-25 00:00:00'),
            'edi_ExchangeListingStatus': 'Delisted',
            'edi_FirstPriceDate': pd.Timestamp('2017-03-01 00:00:00'),
            'edi_GlobalListingStatus': 'Inactive',
            'edi_Industry': 'Real Estate',
            'edi_IsPrimaryListing': True,
            'edi_IsoCountryInc': 'US',
            'edi_IsoCountryListed': 'US',
            'edi_IssuerId': 74929.0,
            'edi_IssuerName': 'Gladstone Commercial Corp',
            'edi_LastPriceDate': pd.Timestamp('2019-10-25 00:00:00'),
            'edi_LocalSymbol': 'GOODO',
            'edi_Mic': 'XNAS',
            'edi_MicSegment': 'XNGS',
            'edi_MicTimezone': 'America/New_York',
            'edi_PreferredName': 'Gladstone Commercial Corp',
            'edi_PrimaryMic': 'XNAS',
            'edi_RecordCreated': pd.Timestamp('2006-10-26 00:00:00'),
            'edi_RecordModified': pd.Timestamp('2019-11-04 03:52:18'),
            'edi_SecId': 426219.0,
            'edi_SecTypeCode': 'PRF',
            'edi_SecTypeDesc': 'Preference Share',
            'edi_SecurityDesc': '7.50% PRF PERPETUAL USD 25 - Reg S Ser B',
            'edi_Sic': 'Lessors of Real Property, Not Elsewhere Classified',
            'edi_SicCode': 6519.0,
            'edi_SicDivision': 'Finance, Insurance, And Real Estate',
            'edi_SicIndustryGroup': 'Real Estate Operators (except Developers) And Lessors',
            'edi_SicMajorGroup': 'Real Estate',
            'edi_StructureCode': 'REIT',
            'edi_StructureDesc': 'Real Estate Investment Trust',
            'ibkr_AggGroup': 'nan',
            'ibkr_Category': 'nan',
            'ibkr_ComboLegs': 'nan',
            'ibkr_ConId': 'nan',
            'ibkr_ContractMonth': 'nan',
            'ibkr_Currency': 'nan',
            'ibkr_Cusip': 'nan',
            'ibkr_DateDelisted': 'nan',
            'ibkr_Delisted': False,
            'ibkr_Etf': False,
            'ibkr_EvMultiplier': 'nan',
            'ibkr_EvRule': 'nan',
            'ibkr_Industry': 'nan',
            'ibkr_Isin': 'nan',
            'ibkr_LastTradeDate': 'nan',
            'ibkr_LocalSymbol': 'nan',
            'ibkr_LongName': 'nan',
            'ibkr_MarketName': 'nan',
            'ibkr_MarketRuleIds': 'nan',
            'ibkr_MdSizeMultiplier': 'nan',
            'ibkr_MinTick': 'nan',
            'ibkr_Multiplier': 'nan',
            'ibkr_PriceMagnifier': 'nan',
            'ibkr_PrimaryExchange': 'nan',
            'ibkr_RealExpirationDate': 'nan',
            'ibkr_Right': 'nan',
            'ibkr_SecType': 'nan',
            'ibkr_Sector': 'nan',
            'ibkr_Strike': 'nan',
            'ibkr_Symbol': 'nan',
            'ibkr_Timezone': 'nan',
            'ibkr_TradingClass': 'nan',
            'ibkr_UnderConId': 'nan',
            'ibkr_UnderSecType': 'nan',
            'ibkr_UnderSymbol': 'nan',
            'ibkr_ValidExchanges': 'nan',
            'sharadar_Category': 'Domestic Preferred Stock',
            'sharadar_CompanySite': 'nan',
            'sharadar_CountryListed': 'US',
            'sharadar_Currency': 'USD',
            'sharadar_Cusips': 376536306.0,
            'sharadar_DateDelisted': pd.Timestamp('2019-10-25 00:00:00'),
            'sharadar_Delisted': True,
            'sharadar_Exchange': 'NASDAQ',
            'sharadar_FamaIndustry': 'Real Estate',
            'sharadar_FamaSector': 'nan',
            'sharadar_FirstAdded': pd.Timestamp('2018-12-30 00:00:00'),
            'sharadar_FirstPriceDate': pd.Timestamp('2006-10-25 00:00:00'),
            'sharadar_FirstQuarter': 'nan',
            'sharadar_Industry': 'REIT - Diversified',
            'sharadar_LastPriceDate': pd.Timestamp('2019-10-25 00:00:00'),
            'sharadar_LastQuarter': 'nan',
            'sharadar_LastUpdated': pd.Timestamp('2020-07-01 00:00:00'),
            'sharadar_Location': 'Virginia; U.S.A',
            'sharadar_Name': 'Gladstone Commercial Corp',
            'sharadar_Permaticker': 112800.0,
            'sharadar_RelatedTickers': 'GOODP GOODM GOOD',
            'sharadar_ScaleMarketCap': 'nan',
            'sharadar_ScaleRevenue': 'nan',
            'sharadar_SecFilings': 'nan',
            'sharadar_Sector': 'Real Estate',
            'sharadar_SicCode': 6519.0,
            'sharadar_SicIndustry': 'Lessors Of Real Property Nec',
            'sharadar_SicSector': 'Finance Insurance And Real Estate',
            'sharadar_Ticker': 'GOODO',
            'usstock_DateDelisted': pd.Timestamp('2019-10-25 00:00:00'),
            'usstock_FirstPriceDate': pd.Timestamp('2007-01-03 00:00:00'),
            'usstock_LastPriceDate': pd.Timestamp('2019-10-25 00:00:00'),
            'usstock_Mic': 'XNAS',
            'usstock_Name': 'GLADSTONE COMMERCIAL COR',
            'usstock_SecurityType': 'PUBLIC',
            'usstock_SecurityType2': 'Preferred Stock',
            'usstock_Sic': 'Lessors of Real Property, Not Elsewhere Classified',
            'usstock_SicCode': 6519.0,
            'usstock_SicDivision': 'Finance, Insurance, And Real Estate',
            'usstock_SicIndustryGroup': 'Real Estate Operators (except Developers) And Lessors',
            'usstock_SicMajorGroup': 'Real Estate',
            'usstock_Symbol': 'GOODO'},
            {'Sid': 'FIBBG000B9XRY4',
            'Symbol': 'AAPL',
            'Exchange': 'XNAS',
            'Country': 'US',
            'Currency': 'USD',
            'SecType': 'STK',
            'Etf': False,
            'Timezone': 'America/New_York',
            'Name': 'APPLE INC',
            'PriceMagnifier': 1,
            'Multiplier': 1,
            'Delisted': False,
            'DateDelisted': 'nan',
            'LastTradeDate': 'nan',
            'RolloverDate': 'nan',
            'alpaca_AssetClass': 'us_equity',
            'alpaca_AssetId': 'b0b6dd9d-8b9b-48a9-ba46-b9d54906e415',
            'alpaca_EasyToBorrow': True,
            'alpaca_Exchange': 'NASDAQ',
            'alpaca_Marginable': True,
            'alpaca_Name': 'Apple Inc. Common Stock',
            'alpaca_Shortable': True,
            'alpaca_Status': 'active',
            'alpaca_Symbol': 'AAPL',
            'alpaca_Tradable': True,
            'edi_Cik': 320193.0,
            'edi_CountryInc': 'United States of America',
            'edi_CountryListed': 'United States of America',
            'edi_Currency': 'USD',
            'edi_DateDelisted': 'nan',
            'edi_ExchangeListingStatus': 'Listed',
            'edi_FirstPriceDate': pd.Timestamp('2017-03-01 00:00:00'),
            'edi_GlobalListingStatus': 'Active',
            'edi_Industry': 'Information Technology',
            'edi_IsPrimaryListing': True,
            'edi_IsoCountryInc': 'US',
            'edi_IsoCountryListed': 'US',
            'edi_IssuerId': 30017.0,
            'edi_IssuerName': 'Apple Inc',
            'edi_LastPriceDate': pd.Timestamp('2020-11-04 00:00:00'),
            'edi_LocalSymbol': 'AAPL',
            'edi_Mic': 'XNAS',
            'edi_MicSegment': 'XNGS',
            'edi_MicTimezone': 'America/New_York',
            'edi_PreferredName': 'Apple Inc',
            'edi_PrimaryMic': 'XNAS',
            'edi_RecordCreated': pd.Timestamp('2001-05-05 00:00:00'),
            'edi_RecordModified': pd.Timestamp('2020-11-02 03:37:23'),
            'edi_SecId': 33449.0,
            'edi_SecTypeCode': 'EQS',
            'edi_SecTypeDesc': 'Equity Shares',
            'edi_SecurityDesc': 'Ordinary Shares',
            'edi_Sic': 'Electronic Computers',
            'edi_SicCode': 3571.0,
            'edi_SicDivision': 'Manufacturing',
            'edi_SicIndustryGroup': 'Computer And Office Equipment',
            'edi_SicMajorGroup': 'Industrial And Commercial Machinery And Computer Equipment',
            'edi_StructureCode': 'nan',
            'edi_StructureDesc': 'nan',
            'ibkr_AggGroup': 1.0,
            'ibkr_Category': 'Computers',
            'ibkr_ComboLegs': 'nan',
            'ibkr_ConId': 265598.0,
            'ibkr_ContractMonth': 'nan',
            'ibkr_Currency': 'USD',
            'ibkr_Cusip': 'nan',
            'ibkr_DateDelisted': 'nan',
            'ibkr_Delisted': False,
            'ibkr_Etf': False,
            'ibkr_EvMultiplier': 0.0,
            'ibkr_EvRule': 'nan',
            'ibkr_Industry': 'Computers',
            'ibkr_Isin': 'US0378331005',
            'ibkr_LastTradeDate': 'nan',
            'ibkr_LocalSymbol': 'AAPL',
            'ibkr_LongName': 'APPLE INC',
            'ibkr_MarketName': 'NMS',
            'ibkr_MarketRuleIds': '26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26,26',
            'ibkr_MdSizeMultiplier': 100.0,
            'ibkr_MinTick': 0.01,
            'ibkr_Multiplier': 'nan',
            'ibkr_PriceMagnifier': 1.0,
            'ibkr_PrimaryExchange': 'NASDAQ',
            'ibkr_RealExpirationDate': 'nan',
            'ibkr_Right': 'nan',
            'ibkr_SecType': 'STK',
            'ibkr_Sector': 'Technology',
            'ibkr_Strike': 0.0,
            'ibkr_Symbol': 'AAPL',
            'ibkr_Timezone': 'America/New_York',
            'ibkr_TradingClass': 'NMS',
            'ibkr_UnderConId': 0.0,
            'ibkr_UnderSecType': 'nan',
            'ibkr_UnderSymbol': 'nan',
            'ibkr_ValidExchanges': 'SMART,AMEX,NYSE,CBOE,PHLX,ISE,CHX,ARCA,ISLAND,DRCTEDGE,BEX,BATS,EDGEA,CSFBALGO,JEFFALGO,BYX,IEX,EDGX,FOXRIVER,TPLUS1,NYSENAT,LTSE,MEMX,PSX',
            'sharadar_Category': 'Domestic Common Stock',
            'sharadar_CompanySite': 'http://www.apple.com',
            'sharadar_CountryListed': 'US',
            'sharadar_Currency': 'USD',
            'sharadar_Cusips': 37833100.0,
            'sharadar_DateDelisted': 'nan',
            'sharadar_Delisted': False,
            'sharadar_Exchange': 'NASDAQ',
            'sharadar_FamaIndustry': 'Computers',
            'sharadar_FamaSector': 'nan',
            'sharadar_FirstAdded': pd.Timestamp('2014-09-24 00:00:00'),
            'sharadar_FirstPriceDate': pd.Timestamp('1986-01-01 00:00:00'),
            'sharadar_FirstQuarter': pd.Timestamp('1996-09-30 00:00:00'),
            'sharadar_Industry': 'Consumer Electronics',
            'sharadar_LastPriceDate': pd.Timestamp('2020-11-04 00:00:00'),
            'sharadar_LastQuarter': pd.Timestamp('2020-09-30 00:00:00'),
            'sharadar_LastUpdated': pd.Timestamp('2020-11-04 00:00:00'),
            'sharadar_Location': 'California; U.S.A',
            'sharadar_Name': 'Apple Inc',
            'sharadar_Permaticker': 199059.0,
            'sharadar_RelatedTickers': 'nan',
            'sharadar_ScaleMarketCap': '6 - Mega',
            'sharadar_ScaleRevenue': '6 - Mega',
            'sharadar_SecFilings': 'https://www.sec.gov/cgi-bin/browse-edgar?action=getcompany&CIK=0000320193',
            'sharadar_Sector': 'Technology',
            'sharadar_SicCode': 3571.0,
            'sharadar_SicIndustry': 'Electronic Computers',
            'sharadar_SicSector': 'Manufacturing',
            'sharadar_Ticker': 'AAPL',
            'usstock_DateDelisted': 'nan',
            'usstock_FirstPriceDate': pd.Timestamp('2007-01-03 00:00:00'),
            'usstock_LastPriceDate': pd.Timestamp('2020-11-18 00:00:00'),
            'usstock_Mic': 'XNAS',
            'usstock_Name': 'APPLE INC',
            'usstock_SecurityType': 'Common Stock',
            'usstock_SecurityType2': 'Common Stock',
            'usstock_Sic': 'Electronic Computers',
            'usstock_SicCode': 3571.0,
            'usstock_SicDivision': 'Manufacturing',
            'usstock_SicIndustryGroup': 'Computer And Office Equipment',
            'usstock_SicMajorGroup': 'Industrial And Commercial Machinery And Computer Equipment',
            'usstock_Symbol': 'AAPL'},
            {'Sid': 'QF000000000004',
            'Symbol': 'ESH7',
            'Exchange': 'XCME',
            'Country': 'US',
            'Currency': 'USD',
            'SecType': 'FUT',
            'Etf': False,
            'Timezone': 'America/Chicago',
            'Name': 'E-mini S&P 500',
            'PriceMagnifier': 1,
            'Multiplier': 50,
            'Delisted': False,
            'DateDelisted': 'nan',
            'LastTradeDate': pd.Timestamp('2017-03-17 00:00:00'),
            'RolloverDate': pd.Timestamp('2017-03-15 00:00:00'),
            'alpaca_AssetClass': 'nan',
            'alpaca_AssetId': 'nan',
            'alpaca_EasyToBorrow': False,
            'alpaca_Exchange': 'nan',
            'alpaca_Marginable': False,
            'alpaca_Name': 'nan',
            'alpaca_Shortable': False,
            'alpaca_Status': 'nan',
            'alpaca_Symbol': 'nan',
            'alpaca_Tradable': False,
            'edi_Cik': 'nan',
            'edi_CountryInc': 'nan',
            'edi_CountryListed': 'nan',
            'edi_Currency': 'nan',
            'edi_DateDelisted': 'nan',
            'edi_ExchangeListingStatus': 'nan',
            'edi_FirstPriceDate': 'nan',
            'edi_GlobalListingStatus': 'nan',
            'edi_Industry': 'nan',
            'edi_IsPrimaryListing': False,
            'edi_IsoCountryInc': 'nan',
            'edi_IsoCountryListed': 'nan',
            'edi_IssuerId': 'nan',
            'edi_IssuerName': 'nan',
            'edi_LastPriceDate': 'nan',
            'edi_LocalSymbol': 'nan',
            'edi_Mic': 'nan',
            'edi_MicSegment': 'nan',
            'edi_MicTimezone': 'nan',
            'edi_PreferredName': 'nan',
            'edi_PrimaryMic': 'nan',
            'edi_RecordCreated': 'nan',
            'edi_RecordModified': 'nan',
            'edi_SecId': 'nan',
            'edi_SecTypeCode': 'nan',
            'edi_SecTypeDesc': 'nan',
            'edi_SecurityDesc': 'nan',
            'edi_Sic': 'nan',
            'edi_SicCode': 'nan',
            'edi_SicDivision': 'nan',
            'edi_SicIndustryGroup': 'nan',
            'edi_SicMajorGroup': 'nan',
            'edi_StructureCode': 'nan',
            'edi_StructureDesc': 'nan',
            'ibkr_AggGroup': 2147483647.0,
            'ibkr_Category': 'nan',
            'ibkr_ComboLegs': 'nan',
            'ibkr_ConId': 215465490.0,
            'ibkr_ContractMonth': 201703.0,
            'ibkr_Currency': 'USD',
            'ibkr_Cusip': 'nan',
            'ibkr_DateDelisted': 'nan',
            'ibkr_Delisted': False,
            'ibkr_Etf': False,
            'ibkr_EvMultiplier': 0.0,
            'ibkr_EvRule': 'nan',
            'ibkr_Industry': 'nan',
            'ibkr_Isin': 'nan',
            'ibkr_LastTradeDate': pd.Timestamp('2017-03-17 00:00:00'),
            'ibkr_LocalSymbol': 'ESH7',
            'ibkr_LongName': 'E-mini S&P 500',
            'ibkr_MarketName': 'ES',
            'ibkr_MarketRuleIds': 'nan',
            'ibkr_MdSizeMultiplier': 1.0,
            'ibkr_MinTick': 0.25,
            'ibkr_Multiplier': 50.0,
            'ibkr_PriceMagnifier': 1.0,
            'ibkr_PrimaryExchange': 'GLOBEX',
            'ibkr_RealExpirationDate': 'nan',
            'ibkr_Right': 'nan',
            'ibkr_SecType': 'FUT',
            'ibkr_Sector': 'nan',
            'ibkr_Strike': 0.0,
            'ibkr_Symbol': 'ES',
            'ibkr_Timezone': 'America/Chicago',
            'ibkr_TradingClass': 'ES',
            'ibkr_UnderConId': 11004968.0,
            'ibkr_UnderSecType': 'IND',
            'ibkr_UnderSymbol': 'ES',
            'ibkr_ValidExchanges': 'GLOBEX',
            'sharadar_Category': 'nan',
            'sharadar_CompanySite': 'nan',
            'sharadar_CountryListed': 'nan',
            'sharadar_Currency': 'nan',
            'sharadar_Cusips': 'nan',
            'sharadar_DateDelisted': 'nan',
            'sharadar_Delisted': False,
            'sharadar_Exchange': 'nan',
            'sharadar_FamaIndustry': 'nan',
            'sharadar_FamaSector': 'nan',
            'sharadar_FirstAdded': 'nan',
            'sharadar_FirstPriceDate': 'nan',
            'sharadar_FirstQuarter': 'nan',
            'sharadar_Industry': 'nan',
            'sharadar_LastPriceDate': 'nan',
            'sharadar_LastQuarter': 'nan',
            'sharadar_LastUpdated': 'nan',
            'sharadar_Location': 'nan',
            'sharadar_Name': 'nan',
            'sharadar_Permaticker': 'nan',
            'sharadar_RelatedTickers': 'nan',
            'sharadar_ScaleMarketCap': 'nan',
            'sharadar_ScaleRevenue': 'nan',
            'sharadar_SecFilings': 'nan',
            'sharadar_Sector': 'nan',
            'sharadar_SicCode': 'nan',
            'sharadar_SicIndustry': 'nan',
            'sharadar_SicSector': 'nan',
            'sharadar_Ticker': 'nan',
            'usstock_DateDelisted': 'nan',
            'usstock_FirstPriceDate': 'nan',
            'usstock_LastPriceDate': 'nan',
            'usstock_Mic': 'nan',
            'usstock_Name': 'nan',
            'usstock_SecurityType': 'nan',
            'usstock_SecurityType2': 'nan',
            'usstock_Sic': 'nan',
            'usstock_SicCode': 'nan',
            'usstock_SicDivision': 'nan',
            'usstock_SicIndustryGroup': 'nan',
            'usstock_SicMajorGroup': 'nan',
            'usstock_Symbol': 'nan'}]
                    )

class SecuritiesReindexedLikeTestCase(unittest.TestCase):

    @patch("quantrocket.master.download_master_file")
    def test_pass_sids_based_on_reindex_like(self, mock_download_master_file):
        """
        Tests that sids are correctly passed to the download_master_file
        function based on reindex_like.
        """
        closes = pd.DataFrame(
            np.random.rand(3,2),
            columns=["FI12345","FI23456"],
            index=pd.date_range(start="2018-05-01", periods=3, freq="D", name="Date"))

        def _mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                          "FI23456"],
                     Symbol=["ABC",
                             "DEF"],
                     Etf=[1,
                          0],
                     Delisted=[0,
                               1],
                     Currency=["USD",
                               "USD"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        mock_download_master_file.side_effect = _mock_download_master_file

        get_securities_reindexed_like(closes, fields=["Symbol", "Etf", "Delisted", "Currency"])

        download_master_file_call = mock_download_master_file.mock_calls[0]
        _, args, kwargs = download_master_file_call
        self.assertListEqual(kwargs["sids"], ["FI12345","FI23456"])
        self.assertListEqual(kwargs["fields"], ["Symbol", "Etf", "Delisted", "Currency"])

    @patch("quantrocket.master.download_master_file")
    def test_cast_boolean_fields(self, mock_download_master_file):
        """
        Tests that master fields Etf and Delisted are cast to boolean.
        """
        closes = pd.DataFrame(
            np.random.rand(3,2),
            columns=["FI12345","FI23456"],
            index=pd.date_range(start="2018-05-01", periods=3, freq="D", name="Date"))

        def _mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(dict(Sid=["FI12345","FI23456"],
                                           Symbol=["ABC","DEF"],
                                           Delisted=[0, 1],
                                           Etf=[1, 0],
                                           edi_Delisted=[0,1],
                                           ibkr_Etf=[1, 0]))
            securities.to_csv(f, index=False)
            f.seek(0)

        mock_download_master_file.side_effect = _mock_download_master_file

        with patch("quantrocket.price.download_master_file", new=mock_download_master_file):

            securities = get_securities_reindexed_like(
                closes,
                fields=["Symbol", "Etf", "Delisted", "edi_Delisted", "ibkr_Etf"])

        symbols = securities.loc["Symbol"]
        symbols = symbols.reset_index()
        symbols.loc[:, "Date"] = symbols.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        self.assertListEqual(
            symbols.to_dict(orient="records"),
            [{'Date': '2018-05-01T00:00:00', "FI12345": "ABC", "FI23456": "DEF"},
            {'Date': '2018-05-02T00:00:00', "FI12345": "ABC", "FI23456": "DEF"},
            {'Date': '2018-05-03T00:00:00', "FI12345": "ABC", "FI23456": "DEF"}]
        )

        for field in ("Delisted", "edi_Delisted"):
            delisted = securities.loc[field]
            delisted = delisted.reset_index()
            delisted.loc[:, "Date"] = delisted.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
            self.assertListEqual(
                delisted.to_dict(orient="records"),
                [{'Date': '2018-05-01T00:00:00', "FI12345": False, "FI23456": True},
                {'Date': '2018-05-02T00:00:00', "FI12345": False, "FI23456": True},
                {'Date': '2018-05-03T00:00:00', "FI12345": False, "FI23456": True}]
            )

        for field in ("Etf", "ibkr_Etf"):
            etfs = securities.loc[field]
            etfs = etfs.reset_index()
            etfs.loc[:, "Date"] = etfs.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")
            self.assertListEqual(
                etfs.to_dict(orient="records"),
                [{'Date': '2018-05-01T00:00:00', "FI12345": True, "FI23456": False},
                {'Date': '2018-05-02T00:00:00', "FI12345": True, "FI23456": False},
                {'Date': '2018-05-03T00:00:00', "FI12345": True, "FI23456": False},
                ]
            )

    def test_securities_reindexed_like(self):
        """
        Tests get_securities_reindexed_like.
        """
        closes = pd.DataFrame(
            np.random.rand(3,2),
            columns=["FI12345","FI23456"],
            index=pd.date_range(start="2018-05-01",
                                periods=3,
                                freq="D",
                                tz="America/New_York",
                                name="Date"))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456"],
                     Symbol=["ABC",
                             "DEF"],
                     Etf=[1,
                          0],
                     Delisted=[0,
                               1],
                     Currency=["USD",
                               "EUR"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            securities = get_securities_reindexed_like(
                closes,
                fields=["Symbol", "Etf", "Delisted", "Currency"])

            self.maxDiff = None

            securities = securities.reset_index()
            securities.loc[:, "Date"] = securities.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")

            self.assertListEqual(
                securities.to_dict(orient="records"),
                [{'Field': 'Currency',
                  'Date': '2018-05-01T00:00:00-0400',
                  "FI12345": 'USD',
                  "FI23456": 'EUR'},
                 {'Field': 'Currency',
                  'Date': '2018-05-02T00:00:00-0400',
                  "FI12345": 'USD',
                  "FI23456": 'EUR'},
                 {'Field': 'Currency',
                  'Date': '2018-05-03T00:00:00-0400',
                  "FI12345": 'USD',
                  "FI23456": 'EUR'},
                 {'Field': 'Delisted',
                  'Date': '2018-05-01T00:00:00-0400',
                  "FI12345": False,
                  "FI23456": True},
                 {'Field': 'Delisted',
                  'Date': '2018-05-02T00:00:00-0400',
                  "FI12345": False,
                  "FI23456": True},
                 {'Field': 'Delisted',
                  'Date': '2018-05-03T00:00:00-0400',
                  "FI12345": False,
                  "FI23456": True},
                 {'Field': 'Etf',
                  'Date': '2018-05-01T00:00:00-0400',
                  "FI12345": True,
                  "FI23456": False},
                 {'Field': 'Etf',
                  'Date': '2018-05-02T00:00:00-0400',
                  "FI12345": True,
                  "FI23456": False},
                 {'Field': 'Etf',
                  'Date': '2018-05-03T00:00:00-0400',
                  "FI12345": True,
                  "FI23456": False},
                 {'Field': 'Symbol',
                  'Date': '2018-05-01T00:00:00-0400',
                  "FI12345": 'ABC',
                  "FI23456": 'DEF'},
                 {'Field': 'Symbol',
                  'Date': '2018-05-02T00:00:00-0400',
                  "FI12345": 'ABC',
                  "FI23456": 'DEF'},
                 {'Field': 'Symbol',
                  'Date': '2018-05-03T00:00:00-0400',
                  "FI12345": 'ABC',
                  "FI23456": 'DEF'}]
            )

    def test_securities_reindexed_like_intraday(self):
        """
        Tests get_securities_reindexed_like with Date and Time in the index.
        """
        closes = pd.DataFrame(
            np.random.rand(4,2),
            columns=["FI12345","FI23456"],
            index=pd.MultiIndex.from_product([
                pd.date_range(start="2018-05-01",
                              periods=2,
                              freq="D",
                              tz="America/New_York",
                              name="Date"),
                ["09:30:00", "09:31:00"],
                ], names=["Date", "Time"]))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456"],
                     Symbol=["ABC",
                             "DEF"]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            securities = get_securities_reindexed_like(
                closes,
                fields="Symbol")

            securities = securities.reset_index()
            securities.loc[:, "Date"] = securities.Date.dt.strftime("%Y-%m-%dT%H:%M:%S%z")

            self.assertListEqual(
                securities.to_dict(orient="records"),
                [{"FI12345": 'ABC',
                  "FI23456": 'DEF',
                  'Date': '2018-05-01T00:00:00-0400',
                  'Field': 'Symbol',
                  'Time': '09:30:00'},
                 {"FI12345": 'ABC',
                  "FI23456": 'DEF',
                  'Date': '2018-05-01T00:00:00-0400',
                  'Field': 'Symbol',
                  'Time': '09:31:00'},
                 {"FI12345": 'ABC',
                  "FI23456": 'DEF',
                  'Date': '2018-05-02T00:00:00-0400',
                  'Field': 'Symbol',
                  'Time': '09:30:00'},
                 {"FI12345": 'ABC',
                  "FI23456": 'DEF',
                  'Date': '2018-05-02T00:00:00-0400',
                  'Field': 'Symbol',
                  'Time': '09:31:00'}]
            )
class ContractNumsReindexedLikeTestCase(unittest.TestCase):

    @patch("quantrocket.master.download_master_file")
    def test_pass_sids_based_on_reindex_like(self, mock_download_master_file):
        """
        Tests that sids are correctly passed to the download_master_file
        function based on reindex_like.
        """
        closes = pd.DataFrame(
            np.random.rand(3,2),
            columns=["FI12345","FI23456"],
            index=pd.date_range(start="2018-05-01", periods=3, freq="D", name="Date"))

        def _mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456"],
                     ibkr_UnderConId=[1,
                                 1],
                     SecType=["FUT",
                              "FUT"],
                     RolloverDate=[
                         "2018-05-02",
                         "2018-07-04"
                     ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        mock_download_master_file.side_effect = _mock_download_master_file

        get_contract_nums_reindexed_like(closes)

        download_master_file_call = mock_download_master_file.mock_calls[0]
        _, args, kwargs = download_master_file_call
        self.assertListEqual(kwargs["sids"], ["FI12345","FI23456"])
        self.assertListEqual(kwargs["fields"], ["RolloverDate", "ibkr_UnderConId", "SecType"])

    def test_complain_if_no_fut(self):
        """
        Tests error handling when you pass a DataFrame with no FUTs.
        """
        closes = pd.DataFrame(
            np.random.rand(3,2),
            columns=["FI12345","FI23456"],
            index=pd.date_range(start="2018-05-01",
                                periods=3,
                                freq="D",
                                tz="America/New_York",
                                name="Date"))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                    dict(Sid=["FI12345",
                                "FI23456"],
                         ibkr_UnderConId=[1,
                                     1],
                         SecType=["STK",
                                  "STK"],
                         RolloverDate=[
                             None,
                             None
                         ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            with self.assertRaises(ParameterError) as cm:
                get_contract_nums_reindexed_like(closes)

        self.assertIn("input DataFrame does not appear to contain any futures contracts",
                      repr(cm.exception))

    def test_contract_nums_reindexed_like(self):
        """
        Tests get_contract_nums_reindexed_like.
        """
        closes = pd.DataFrame(
            np.random.rand(4,6),
            columns=["FI12345","FI23456", "FI34567", "FI45678", "FI56789", "FI67890"],
            index=pd.date_range(start="2018-05-01",
                                periods=4,
                                freq="D",
                                tz="America/New_York",
                                name="Date"))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456",
                            "FI34567",
                            "FI45678",
                            "FI56789",
                            "FI67890"],
                     ibkr_UnderConId=[1,
                                 2,
                                 1,
                                 2,
                                 None,
                                 2],
                     SecType=["FUT",
                              "FUT",
                              "FUT",
                              "FUT",
                              "STK",
                              "FUT"],
                     RolloverDate=[
                         "2018-05-02",
                         "2018-06-02",
                         "2018-06-03",
                         "2018-05-03",
                         None,
                         "2018-07-03"
                     ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            contract_nums = get_contract_nums_reindexed_like(closes)

        contract_nums.index = contract_nums.index.strftime("%Y-%m-%d")
        contract_nums = contract_nums.fillna("nan")
        self.assertDictEqual(
            contract_nums.to_dict(orient="index"),
            {'2018-05-01': {
                "FI12345": 1.0,
                "FI23456": 2.0,
                "FI34567": 2.0,
                "FI45678": 1.0,
                "FI56789": 'nan',
                "FI67890": 3.0},
             '2018-05-02': {
                 "FI12345": 1.0,
                 "FI23456": 2.0,
                 "FI34567": 2.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 3.0
                },
             '2018-05-03': {
                 "FI12345": 'nan',
                 "FI23456": 2.0,
                 "FI34567": 1.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 3.0},
             '2018-05-04': {
                 "FI12345": 'nan',
                 "FI23456": 1.0,
                 "FI34567": 1.0,
                 "FI45678": 'nan',
                 "FI56789": 'nan',
                 "FI67890": 2.0}}
        )

    def test_contract_nums_reindexed_like_tz_naive(self):
        """
        Tests get_contract_nums_reindexed_like with a tz-naive reindex_like.
        """
        closes = pd.DataFrame(
            np.random.rand(4,6),
            columns=["FI12345","FI23456", "FI34567", "FI45678", "FI56789", "FI67890"],
            index=pd.date_range(start="2018-05-01",
                                periods=4,
                                freq="D",
                                name="Date"))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456",
                            "FI34567",
                            "FI45678",
                            "FI56789",
                            "FI67890"],
                     ibkr_UnderConId=[1,
                                 2,
                                 1,
                                 2,
                                 None,
                                 2],
                     SecType=["FUT",
                              "FUT",
                              "FUT",
                              "FUT",
                              "STK",
                              "FUT"],
                     RolloverDate=[
                         "2018-05-02",
                         "2018-06-02",
                         "2018-06-03",
                         "2018-05-03",
                         None,
                         "2018-07-03"
                     ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            contract_nums = get_contract_nums_reindexed_like(closes)

        contract_nums.index = contract_nums.index.strftime("%Y-%m-%d")
        contract_nums = contract_nums.fillna("nan")
        self.assertDictEqual(
            contract_nums.to_dict(orient="index"),
            {'2018-05-01': {
                "FI12345": 1.0,
                "FI23456": 2.0,
                "FI34567": 2.0,
                "FI45678": 1.0,
                "FI56789": 'nan',
                "FI67890": 3.0},
             '2018-05-02': {
                 "FI12345": 1.0,
                 "FI23456": 2.0,
                 "FI34567": 2.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 3.0
                },
             '2018-05-03': {
                 "FI12345": 'nan',
                 "FI23456": 2.0,
                 "FI34567": 1.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 3.0},
             '2018-05-04': {
                 "FI12345": 'nan',
                 "FI23456": 1.0,
                 "FI34567": 1.0,
                 "FI45678": 'nan',
                 "FI56789": 'nan',
                 "FI67890": 2.0}}
        )

    def test_limit_sequence_num(self):
        """
        Tests get_contract_nums_reindexed_like with the limit option.
        """
        closes = pd.DataFrame(
            np.random.rand(4,6),
            columns=["FI12345","FI23456", "FI34567", "FI45678", "FI56789", "FI67890"],
            index=pd.date_range(start="2018-05-01",
                                periods=4,
                                freq="D",
                                tz="America/New_York",
                                name="Date"))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456",
                            "FI34567",
                            "FI45678",
                            "FI56789",
                            "FI67890"],
                     ibkr_UnderConId=[1,
                                 2,
                                 1,
                                 2,
                                 None,
                                 2],
                     SecType=["FUT",
                              "FUT",
                              "FUT",
                              "FUT",
                              "STK",
                              "FUT"],
                     RolloverDate=[
                         "2018-05-02",
                         "2018-06-02",
                         "2018-06-03",
                         "2018-05-03",
                         None,
                         "2018-07-03"
                     ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            contract_nums = get_contract_nums_reindexed_like(closes, limit=2)

        contract_nums.index = contract_nums.index.strftime("%Y-%m-%d")
        contract_nums = contract_nums.fillna("nan")
        self.assertDictEqual(
            contract_nums.to_dict(orient="index"),
            {'2018-05-01': {
                "FI12345": 1.0,
                "FI23456": 2.0,
                "FI34567": 2.0,
                "FI45678": 1.0,
                "FI56789": 'nan',
                "FI67890": 'nan'},
             '2018-05-02': {
                 "FI12345": 1.0,
                 "FI23456": 2.0,
                 "FI34567": 2.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 'nan'
                },
             '2018-05-03': {
                 "FI12345": 'nan',
                 "FI23456": 2.0,
                 "FI34567": 1.0,
                 "FI45678": 1.0,
                 "FI56789": 'nan',
                 "FI67890": 'nan'},
             '2018-05-04': {
                 "FI12345": 'nan',
                 "FI23456": 1.0,
                 "FI34567": 1.0,
                 "FI45678": 'nan',
                 "FI56789": 'nan',
                 "FI67890": 2.0}}
        )

    def test_contract_nums_reindexed_like_intraday(self):
        """
        Tests get_contract_nums_reindexed_like when the input DataFrame includes
        Dates and Times.
        """
        closes = pd.DataFrame(
            np.random.rand(8,6),
            columns=["FI12345","FI23456", "FI34567", "FI45678", "FI56789", "FI67890"],
            index=pd.MultiIndex.from_product([
                pd.date_range(start="2018-05-01",
                              periods=4,
                              freq="D",
                              tz="America/New_York",
                              name="Date"),
                ["09:30:00","09:31:00"],
            ], names=["Date","Time"]))

        def mock_download_master_file(f, *args, **kwargs):
            securities = pd.DataFrame(
                dict(Sid=["FI12345",
                            "FI23456",
                            "FI34567",
                            "FI45678",
                            "FI56789",
                            "FI67890"],
                     ibkr_UnderConId=[1,
                                 2,
                                 1,
                                 2,
                                 None,
                                 2],
                     SecType=["FUT",
                              "FUT",
                              "FUT",
                              "FUT",
                              "STK",
                              "FUT"],
                     RolloverDate=[
                         "2018-05-02",
                         "2018-06-02",
                         "2018-06-03",
                         "2018-05-03",
                         None,
                         "2018-07-03"
                     ]))
            securities.to_csv(f, index=False)
            f.seek(0)

        with patch('quantrocket.master.download_master_file', new=mock_download_master_file):

            contract_nums = get_contract_nums_reindexed_like(closes)

        self.assertEqual(list(contract_nums.index.names), ["Date","Time"])

        contract_nums = contract_nums.reset_index().fillna("nan")
        contract_nums.loc[:, "Date"] = contract_nums.Date.dt.strftime("%Y-%m-%d")

        self.assertListEqual(
            contract_nums.to_dict(orient="records"),
            [{"FI12345": 1.0,
              "FI23456": 2.0,
              "FI34567": 2.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-01',
              'Time': '09:30:00'},
             {"FI12345": 1.0,
              "FI23456": 2.0,
              "FI34567": 2.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-01',
              'Time': '09:31:00'},
             {"FI12345": 1.0,
              "FI23456": 2.0,
              "FI34567": 2.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-02',
              'Time': '09:30:00'},
             {"FI12345": 1.0,
              "FI23456": 2.0,
              "FI34567": 2.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-02',
              'Time': '09:31:00'},
             {"FI12345": 'nan',
              "FI23456": 2.0,
              "FI34567": 1.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-03',
              'Time': '09:30:00'},
             {"FI12345": 'nan',
              "FI23456": 2.0,
              "FI34567": 1.0,
              "FI45678": 1.0,
              "FI56789": 'nan',
              "FI67890": 3.0,
              'Date': '2018-05-03',
              'Time': '09:31:00'},
             {"FI12345": 'nan',
              "FI23456": 1.0,
              "FI34567": 1.0,
              "FI45678": 'nan',
              "FI56789": 'nan',
              "FI67890": 2.0,
              'Date': '2018-05-04',
              'Time': '09:30:00'},
             {"FI12345": 'nan',
              "FI23456": 1.0,
              "FI34567": 1.0,
              "FI45678": 'nan',
              "FI56789": 'nan',
              "FI67890": 2.0,
              'Date': '2018-05-04',
              'Time': '09:31:00'}]
        )
